# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackSkill, LevelGenerator, RewardManager
from minihack.envs import register


levitation_msg = [
    "You float up",
    "You start to float in the air",
    "Up, up, and awaaaay!",
    "a ring of levitation (on left hand)",
    "a ring of levitation (on right hand)",
]


class MiniHackLevitate(MiniHackSkill):
    def __init__(self, *args, des_file, **kwargs):
        rwrd_mngr = RewardManager()
        rwrd_mngr.add_message_event(levitation_msg)

        super().__init__(
            *args, des_file=des_file, reward_manager=rwrd_mngr, **kwargs
        )


class MiniHackLevitateBoots(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("levitation boots", "[", cursestate="blessed")
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLevitateBootsFixed(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object(
            "levitation boots", "[", place=(0, 0), cursestate="blessed"
        )
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLevitateRing(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("levitation", "=", cursestate="blessed")
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLevitateRingFixed(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object(
            "levitation", "=", place=(0, 0), cursestate="blessed"
        )
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLevitatePotion(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("levitation", "!", cursestate="blessed")
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLevitatePotionFixed(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object(
            "levitation", "!", place=(0, 0), cursestate="blessed"
        )
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackLevitateRandom(MiniHackLevitate):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
.....
.....
.....
.....
.....
ENDMAP
REGION:(0,0,5,5),lit,"ordinary"
IF [33%] {
    OBJECT:('!',"levitation"),random,blessed
} ELSE {
    IF [50%] {
        OBJECT:('=',"levitation"),random,blessed
    } ELSE {
        OBJECT:('[',"levitation boots"),random,blessed
    }
}
"""
        super().__init__(*args, des_file=des_file, **kwargs)


register(
    id="MiniHack-Levitate-Boots-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitateBoots",
)
register(
    id="MiniHack-Levitate-Ring-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitateRing",
)
register(
    id="MiniHack-Levitate-Potion-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitatePotion",
)
register(
    id="MiniHack-Levitate-Random-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitateRandom",
)
register(
    id="MiniHack-Levitate-Boots-Fixed-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitateBootsFixed",
)
register(
    id="MiniHack-Levitate-Ring-Fixed-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitateRingFixed",
)
register(
    id="MiniHack-Levitate-Potion-Fixed-v0",
    entry_point="minihack.envs.skills_levitate:MiniHackLevitatePotionFixed",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Installation for hydra:
pip install hydra-core hydra_colorlog --upgrade

Runs like polybeast but use = to set flags:
python -m polyhydra.py learning_rate=0.001 rnd.twoheaded=true

Run sweep with another -m after the module:
python -m polyhydra.py -m learning_rate=0.01,0.001,0.0001,0.00001 momentum=0,0.5
"""

import hydra
from omegaconf import OmegaConf, DictConfig
from pathlib import Path

import logging
import os

import numpy as np
import multiprocessing as mp

from minihack.agent import is_env_registered, get_env_shortcut
from minihack.agent.polybeast import polybeast_env, polybeast_learner
from minihack.agent.common.envs import tasks

import torch

if torch.__version__.startswith("1.5") or torch.__version__.startswith("1.6"):
    # pytorch 1.5.* needs this for some reason on the cluster
    os.environ["MKL_SERVICE_FORCE_INTEL"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

logging.basicConfig(
    format=(
        "[%(levelname)s:%(process)d %(module)s:%(lineno)d %(asctime)s] "
        "%(message)s"
    ),
    level=0,
)


def pipes_basename():
    logdir = Path(os.getcwd())
    name = ".".join(
        [logdir.parents[1].name, logdir.parents[0].name, logdir.name]
    )
    return "unix:/tmp/poly.%s" % name


def get_common_flags(flags):
    flags = OmegaConf.to_container(flags)
    flags["pipes_basename"] = pipes_basename()
    flags["savedir"] = os.getcwd()
    return OmegaConf.create(flags)


def get_learner_flags(flags):
    lrn_flags = OmegaConf.to_container(flags)
    lrn_flags["checkpoint"] = os.path.join(flags["savedir"], "checkpoint.tar")
    lrn_flags["entropy_cost"] = float(lrn_flags["entropy_cost"])
    return OmegaConf.create(lrn_flags)


def run_learner(flags: DictConfig):
    polybeast_learner.main(flags)


def get_environment_flags(flags):
    env_flags = OmegaConf.to_container(flags)
    env_flags["num_servers"] = flags.num_actors
    env_flags["seedspath"] = ""
    return OmegaConf.create(env_flags)


def run_env(flags):
    np.random.seed()  # Get new random seed in forked process.
    polybeast_env.main(flags)


def symlink_latest(savedir, symlink):
    try:
        if os.path.islink(symlink):
            os.remove(symlink)
        if not os.path.exists(symlink):
            os.symlink(savedir, symlink)
            logging.info("Symlinked log directory: %s" % symlink)
    except OSError:
        # os.remove() or os.symlink() raced. Don't do anything.
        pass


@hydra.main(config_path=".", config_name="config")
def main(flags: DictConfig):
    if os.path.exists("config.yaml"):
        # this ignores the local config.yaml and replaces it completely with saved one
        logging.info(
            "loading existing configuration, we're continuing a previous run"
        )
        new_flags = OmegaConf.load("config.yaml")
        cli_conf = OmegaConf.from_cli()
        # however, you can override parameters from the cli still
        # this is useful e.g. if you did total_steps=N before and want to increase it
        flags = OmegaConf.merge(new_flags, cli_conf)

    logging.info(OmegaConf.to_yaml(flags))
    OmegaConf.save(flags, "config.yaml")

    flags = get_common_flags(flags)

    # check the name of the environment
    if flags.env not in tasks.ENVS:
        if is_env_registered(flags.env):
            flags.env = get_env_shortcut(flags.env)
        else:
            raise KeyError(
                f"Could not find an environement with a name: {flags.env}."
            )

    # set flags for polybeast_env
    env_flags = get_environment_flags(flags)
    env_processes = []
    for _ in range(1):
        p = mp.Process(target=run_env, args=(env_flags,))
        p.start()
        env_processes.append(p)

    symlink_latest(
        flags.savedir, os.path.join(hydra.utils.get_original_cwd(), "latest")
    )

    lrn_flags = get_learner_flags(flags)
    run_learner(lrn_flags)

    for p in env_processes:
        p.kill()
        p.join()


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import csv
import datetime
import json
import logging
import os
import time
import weakref


def _save_metadata(path, metadata):
    metadata["date_save"] = datetime.datetime.now().strftime(
        "%Y-%m-%d %H:%M:%S.%f"
    )
    with open(path, "w") as f:
        json.dump(metadata, f, indent=4, sort_keys=True)


def gather_metadata():
    metadata = dict(
        date_start=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
        env=os.environ.copy(),
        successful=False,
    )

    # Git metadata.
    try:
        import git
    except ImportError:
        logging.warning(
            "Couldn't import gitpython module; install it with `pip install gitpython`."
        )
    else:
        try:
            repo = git.Repo(search_parent_directories=True)
            metadata["git"] = {
                "commit": repo.commit().hexsha,
                "is_dirty": repo.is_dirty(),
                "path": repo.git_dir,
            }
            if not repo.head.is_detached:
                metadata["git"]["branch"] = repo.active_branch.name
        except git.InvalidGitRepositoryError:
            pass

    if "git" not in metadata:
        logging.warning("Couldn't determine git data.")

    # Slurm metadata.
    if "SLURM_JOB_ID" in os.environ:
        slurm_env_keys = [k for k in os.environ if k.startswith("SLURM")]
        metadata["slurm"] = {}
        for k in slurm_env_keys:
            d_key = k.replace("SLURM_", "").replace("SLURMD_", "").lower()
            metadata["slurm"][d_key] = os.environ[k]

    return metadata


class FileWriter:
    def __init__(self, xp_args=None, rootdir="~/palaas"):
        if rootdir == "~/palaas":
            # make unique id in case someone uses the default rootdir
            xpid = "{proc}_{unixtime}".format(
                proc=os.getpid(), unixtime=int(time.time())
            )
            rootdir = os.path.join(rootdir, xpid)
        self.basepath = os.path.expandvars(os.path.expanduser(rootdir))

        self._tick = 0

        # metadata gathering
        if xp_args is None:
            xp_args = {}
        self.metadata = gather_metadata()
        # we need to copy the args, otherwise when we close the file writer
        # (and rewrite the args) we might have non-serializable objects (or
        # other nasty stuff).
        self.metadata["args"] = copy.deepcopy(xp_args)

        formatter = logging.Formatter("%(message)s")
        self._logger = logging.getLogger("palaas/out")

        # to stdout handler
        shandle = logging.StreamHandler()
        shandle.setFormatter(formatter)
        self._logger.addHandler(shandle)
        self._logger.setLevel(logging.INFO)

        # to file handler
        if not os.path.exists(self.basepath):
            self._logger.info("Creating log directory: %s", self.basepath)
            os.makedirs(self.basepath, exist_ok=True)
        else:
            self._logger.info("Found log directory: %s", self.basepath)

        self.paths = dict(
            msg="{base}/out.log".format(base=self.basepath),
            logs="{base}/logs.csv".format(base=self.basepath),
            fields="{base}/fields.csv".format(base=self.basepath),
            meta="{base}/meta.json".format(base=self.basepath),
        )

        self._logger.info("Saving arguments to %s", self.paths["meta"])
        if os.path.exists(self.paths["meta"]):
            self._logger.warning(
                "Path to meta file already exists. " "Not overriding meta."
            )
        else:
            self.save_metadata()

        self._logger.info("Saving messages to %s", self.paths["msg"])
        if os.path.exists(self.paths["msg"]):
            self._logger.warning(
                "Path to message file already exists. "
                "New data will be appended."
            )

        fhandle = logging.FileHandler(self.paths["msg"])
        fhandle.setFormatter(formatter)
        self._logger.addHandler(fhandle)

        self._logger.info("Saving logs data to %s", self.paths["logs"])
        self._logger.info("Saving logs' fields to %s", self.paths["fields"])
        self.fieldnames = ["_tick", "_time"]
        if os.path.exists(self.paths["logs"]):
            self._logger.warning(
                "Path to log file already exists. "
                "New data will be appended."
            )
            # Override default fieldnames.
            with open(self.paths["fields"], "r") as csvfile:
                reader = csv.reader(csvfile)
                lines = list(reader)
                if len(lines) > 0:
                    self.fieldnames = lines[-1]
            # Override default tick: use the last tick from the logs file plus 1.
            with open(self.paths["logs"], "r") as csvfile:
                reader = csv.reader(csvfile)
                lines = list(reader)
                # Need at least two lines in order to read the last tick:
                # the first is the csv header and the second is the first line
                # of data.
                if len(lines) > 1:
                    self._tick = int(lines[-1][0]) + 1

        self._fieldfile = open(self.paths["fields"], "a")
        self._fieldwriter = csv.writer(self._fieldfile)
        self._fieldfile.flush()
        self._logfile = open(self.paths["logs"], "a")
        self._logwriter = csv.DictWriter(
            self._logfile, fieldnames=self.fieldnames
        )

        # Auto-close (and save) on destruction.
        weakref.finalize(
            self, _save_metadata, self.paths["meta"], self.metadata
        )

    def log(self, to_log, tick=None, verbose=False):
        if tick is not None:
            raise NotImplementedError
        else:
            to_log["_tick"] = self._tick
            self._tick += 1
        to_log["_time"] = time.time()

        old_len = len(self.fieldnames)
        for k in to_log:
            if k not in self.fieldnames:
                self.fieldnames.append(k)
        if old_len != len(self.fieldnames):
            self._fieldwriter.writerow(self.fieldnames)
            self._fieldfile.flush()
            self._logger.info("Updated log fields: %s", self.fieldnames)

        if to_log["_tick"] == 0:
            self._logfile.write("# %s\n" % ",".join(self.fieldnames))

        if verbose:
            self._logger.info(
                "LOG | %s",
                ", ".join(
                    ["{}: {}".format(k, to_log[k]) for k in sorted(to_log)]
                ),
            )

        self._logwriter.writerow(to_log)
        self._logfile.flush()

    def close(self, successful=True):
        self.metadata["successful"] = successful
        self.save_metadata()

        for f in [self._logfile, self._fieldfile]:
            f.close()

    def save_metadata(self):
        _save_metadata(self.paths["meta"], self.metadata)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import torch
from torch import nn
from torch.nn import functional as F

from nle import nethack

from minihack.agent.common.models.embed import GlyphEmbedding
from minihack.agent.common.models.transformer import TransformerEncoder

NUM_GLYPHS = nethack.MAX_GLYPH
NUM_FEATURES = nethack.BLSTATS_SHAPE[0]
PAD_CHAR = 0
NUM_CHARS = 256


class NetHackNet(nn.Module):
    AgentOutput = collections.namedtuple(
        "AgentOutput", "action policy_logits baseline"
    )

    def __init__(self):
        super(NetHackNet, self).__init__()

        self.register_buffer("reward_sum", torch.zeros(()))
        self.register_buffer("reward_m2", torch.zeros(()))
        self.register_buffer("reward_count", torch.zeros(()).fill_(1e-8))

    def forward(self, inputs, core_state):
        raise NotImplementedError

    def initial_state(self, batch_size=1):
        return ()

    def prepare_input(self, inputs):
        # -- [T x B x H x W]
        glyphs = inputs["glyphs"]

        # -- [T x B x F]
        features = inputs["blstats"]

        T, B, *_ = glyphs.shape

        # -- [B' x H x W]
        glyphs = torch.flatten(glyphs, 0, 1)  # Merge time and batch.

        # -- [B' x F]
        features = features.view(T * B, -1).float()

        return glyphs, features

    def embed_state(self, inputs):
        raise NotImplementedError

    @torch.no_grad()
    def update_running_moments(self, reward_batch):
        """Maintains a running mean of reward."""
        new_count = len(reward_batch)
        new_sum = torch.sum(reward_batch)
        new_mean = new_sum / new_count

        curr_mean = self.reward_sum / self.reward_count
        new_m2 = torch.sum((reward_batch - new_mean) ** 2) + (
            (self.reward_count * new_count)
            / (self.reward_count + new_count)
            * (new_mean - curr_mean) ** 2
        )

        self.reward_count += new_count
        self.reward_sum += new_sum
        self.reward_m2 += new_m2

    @torch.no_grad()
    def get_running_std(self):
        """Returns standard deviation of the running mean of the reward."""
        return torch.sqrt(self.reward_m2 / self.reward_count)


class RandomNet(NetHackNet):
    def __init__(self, observation_shape, num_actions, flags, device=None):
        super(RandomNet, self).__init__()
        self.num_actions = num_actions
        self.theta = torch.nn.Parameter(torch.zeros(self.num_actions))

    def forward(self, inputs, core_state):
        T, B, *_ = inputs["glyphs"].shape
        zeros = self.theta * 0
        # set logits to 0
        policy_logits = zeros[None, :].expand(T * B, -1)
        # set baseline to 0
        baseline = policy_logits.sum(dim=1).view(-1, B)

        # sample random action
        action = torch.multinomial(
            F.softmax(policy_logits, dim=1), num_samples=1
        ).view(T, B)
        policy_logits = policy_logits.view(T, B, self.num_actions)
        return (
            dict(
                policy_logits=policy_logits, baseline=baseline, action=action
            ),
            core_state,
        )

    def embed_state(self, inputs):
        raise NotImplementedError


class Crop(nn.Module):
    def __init__(
        self, height, width, height_target, width_target, device=None
    ):
        super(Crop, self).__init__()
        self.width = width
        self.height = height
        self.width_target = width_target
        self.height_target = height_target
        self.width_grid = self._step_to_range(
            2 / (self.width - 1), self.width_target
        )[None, :].expand(self.height_target, -1)
        self.height_grid = self._step_to_range(
            2 / (self.height - 1), height_target
        )[:, None].expand(-1, self.width_target)

        if device is not None:
            self.width_grid = self.width_grid.to(device)
            self.height_grid = self.height_grid.to(device)

    def _step_to_range(self, step, num_steps):
        return torch.tensor(
            [step * (i - num_steps // 2) for i in range(num_steps)]
        )

    def forward(self, inputs, coordinates):
        """Calculates centered crop around given x,y coordinates.

        Args:
           inputs [B x H x W] or [B x H x W x C]
           coordinates [B x 2] x,y coordinates

        Returns:
           [B x H' x W'] inputs cropped and centered around x,y coordinates.
        """
        assert inputs.shape[1] == self.height, "expected %d but found %d" % (
            self.height,
            inputs.shape[1],
        )
        assert inputs.shape[2] == self.width, "expected %d but found %d" % (
            self.width,
            inputs.shape[2],
        )

        permute_results = False
        if inputs.dim() == 3:
            inputs = inputs.unsqueeze(1)
        else:
            permute_results = True
            inputs = inputs.permute(0, 2, 3, 1)
        inputs = inputs.float()

        x = coordinates[:, 0]
        y = coordinates[:, 1]

        x_shift = 2 / (self.width - 1) * (x.float() - self.width // 2)
        y_shift = 2 / (self.height - 1) * (y.float() - self.height // 2)

        grid = torch.stack(
            [
                self.width_grid[None, :, :] + x_shift[:, None, None],
                self.height_grid[None, :, :] + y_shift[:, None, None],
            ],
            dim=3,
        )

        crop = (
            torch.round(F.grid_sample(inputs, grid, align_corners=True))
            .squeeze(1)
            .long()
        )

        if permute_results:
            # [B x C x H x W] -> [B x H x W x C]
            crop = crop.permute(0, 2, 3, 1)

        return crop


class Flatten(nn.Module):
    def forward(self, input):
        return input.view(input.size(0), -1)


class BaseNet(NetHackNet):
    def __init__(self, observation_shape, num_actions, flags, device):
        super(BaseNet, self).__init__()

        self.flags = flags

        self.observation_shape = observation_shape

        self.H = observation_shape[0]
        self.W = observation_shape[1]

        self.num_actions = num_actions
        self.use_lstm = flags.use_lstm

        self.k_dim = flags.embedding_dim
        self.h_dim = flags.hidden_dim

        self.crop_model = flags.crop_model
        self.crop_dim = flags.crop_dim

        self.num_features = NUM_FEATURES

        self.crop = Crop(self.H, self.W, self.crop_dim, self.crop_dim, device)

        self.glyph_type = flags.glyph_type
        self.glyph_embedding = GlyphEmbedding(
            flags.glyph_type,
            flags.embedding_dim,
            device,
            flags.use_index_select,
        )

        K = flags.embedding_dim  # number of input filters
        F = 3  # filter dimensions
        S = 1  # stride
        P = 1  # padding
        M = 16  # number of intermediate filters
        self.Y = 8  # number of output filters
        L = flags.layers  # number of convnet layers

        in_channels = [K] + [M] * (L - 1)
        out_channels = [M] * (L - 1) + [self.Y]

        def interleave(xs, ys):
            return [val for pair in zip(xs, ys) for val in pair]

        conv_extract = [
            nn.Conv2d(
                in_channels=in_channels[i],
                out_channels=out_channels[i],
                kernel_size=(F, F),
                stride=S,
                padding=P,
            )
            for i in range(L)
        ]

        self.extract_representation = nn.Sequential(
            *interleave(conv_extract, [nn.ELU()] * len(conv_extract))
        )

        if self.crop_model == "transformer":
            self.extract_crop_representation = TransformerEncoder(
                K,
                N=L,
                heads=8,
                height=self.crop_dim,
                width=self.crop_dim,
                device=device,
            )
        elif self.crop_model == "cnn":
            conv_extract_crop = [
                nn.Conv2d(
                    in_channels=in_channels[i],
                    out_channels=out_channels[i],
                    kernel_size=(F, F),
                    stride=S,
                    padding=P,
                )
                for i in range(L)
            ]

            self.extract_crop_representation = nn.Sequential(
                *interleave(conv_extract_crop, [nn.ELU()] * len(conv_extract))
            )

        # MESSAGING MODEL
        if "msg" not in flags:
            self.msg_model = "none"
        else:
            self.msg_model = flags.msg.model
            self.msg_hdim = flags.msg.hidden_dim
            self.msg_edim = flags.msg.embedding_dim
        if self.msg_model in ("gru", "lstm", "lt_cnn"):
            # character-based embeddings
            self.char_lt = nn.Embedding(
                NUM_CHARS, self.msg_edim, padding_idx=PAD_CHAR
            )
        else:
            # forward will set up one-hot inputs for the cnn, no lt needed
            pass

        if self.msg_model.endswith("cnn"):
            # from Zhang et al, 2016
            # Character-level Convolutional Networks for Text Classification
            # https://arxiv.org/abs/1509.01626
            if self.msg_model == "cnn":
                # inputs will be one-hot vectors, as done in paper
                self.conv1 = nn.Conv1d(NUM_CHARS, self.msg_hdim, kernel_size=7)
            elif self.msg_model == "lt_cnn":
                # replace one-hot inputs with learned embeddings
                self.conv1 = nn.Conv1d(
                    self.msg_edim, self.msg_hdim, kernel_size=7
                )
            else:
                raise NotImplementedError("msg.model == %s", flags.msg.model)

            # remaining convolutions, relus, pools, and a small FC network
            self.conv2_6_fc = nn.Sequential(
                nn.ReLU(),
                nn.MaxPool1d(kernel_size=3, stride=3),
                # conv2
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=7),
                nn.ReLU(),
                nn.MaxPool1d(kernel_size=3, stride=3),
                # conv3
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                # conv4
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                # conv5
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                # conv6
                nn.Conv1d(self.msg_hdim, self.msg_hdim, kernel_size=3),
                nn.ReLU(),
                nn.MaxPool1d(kernel_size=3, stride=3),
                # fc receives -- [ B x h_dim x 5 ]
                Flatten(),
                nn.Linear(5 * self.msg_hdim, 2 * self.msg_hdim),
                nn.ReLU(),
                nn.Linear(2 * self.msg_hdim, self.msg_hdim),
            )  # final output -- [ B x h_dim x 5 ]
        elif self.msg_model in ("gru", "lstm"):

            def rnn(flag):
                return nn.LSTM if flag == "lstm" else nn.GRU

            self.char_rnn = rnn(self.msg_model)(
                self.msg_edim,
                self.msg_hdim // 2,
                batch_first=True,
                bidirectional=True,
            )
        elif self.msg_model != "none":
            raise NotImplementedError("msg.model == %s", flags.msg.model)

        self.embed_features = nn.Sequential(
            nn.Linear(self.num_features, self.k_dim),
            nn.ReLU(),
            nn.Linear(self.k_dim, self.k_dim),
            nn.ReLU(),
        )

        self.equalize_input_dim = flags.equalize_input_dim
        if not self.equalize_input_dim:
            # just added up the output dimensions of the input featurizers
            # feature / status dim
            out_dim = self.k_dim
            # CNN over full glyph map
            out_dim += self.H * self.W * self.Y
            if self.crop_model == "transformer":
                out_dim += self.crop_dim ** 2 * K
            elif self.crop_model == "cnn":
                out_dim += self.crop_dim ** 2 * self.Y
            # messaging model
            if self.msg_model != "none":
                out_dim += self.msg_hdim
        else:
            # otherwise, project them all to h_dim
            NUM_INPUTS = 4 if self.msg_model != "none" else 3
            project_hdim = flags.equalize_factor * self.h_dim
            out_dim = project_hdim * NUM_INPUTS

            # set up linear layers for projections
            self.project_feature_dim = nn.Linear(self.k_dim, project_hdim)
            self.project_glyph_dim = nn.Linear(
                self.H * self.W * self.Y, project_hdim
            )
            c__2 = self.crop_dim ** 2
            if self.crop_model == "transformer":
                self.project_crop_dim = nn.Linear(c__2 * K, project_hdim)
            elif self.crop_model == "cnn":
                self.project_crop_dim = nn.Linear(c__2 * self.Y, project_hdim)
            if self.msg_model != "none":
                self.project_msg_dim = nn.Linear(self.msg_hdim, project_hdim)

        self.fc = nn.Sequential(
            nn.Linear(out_dim, self.h_dim),
            nn.ReLU(),
            nn.Linear(self.h_dim, self.h_dim),
            nn.ReLU(),
        )

        if self.use_lstm:
            self.core = nn.LSTM(self.h_dim, self.h_dim, num_layers=1)

        self.policy = nn.Linear(self.h_dim, self.num_actions)
        self.baseline = nn.Linear(self.h_dim, 1)

    def initial_state(self, batch_size=1):
        if not self.use_lstm:
            return tuple()
        return tuple(
            torch.zeros(
                self.core.num_layers, batch_size, self.core.hidden_size
            )
            for _ in range(2)
        )

    def prepare_input(self, inputs):
        # -- [T x B x H x W]
        T, B, H, W = inputs["glyphs"].shape

        # take our chosen glyphs and merge the time and batch

        glyphs = self.glyph_embedding.prepare_input(inputs)

        # -- [T x B x F]
        features = inputs["blstats"]
        # -- [B' x F]
        features = features.view(T * B, -1).float()

        return glyphs, features

    def forward(self, inputs, core_state, learning=False):
        T, B, *_ = inputs["glyphs"].shape

        glyphs, features = self.prepare_input(inputs)

        # -- [B x 2] x,y coordinates
        coordinates = features[:, :2]

        features = features.view(T * B, -1).float()
        # -- [B x K]
        features_emb = self.embed_features(features)
        if self.equalize_input_dim:
            features_emb = self.project_feature_dim(features_emb)

        assert features_emb.shape[0] == T * B

        reps = [features_emb]

        # -- [B x H' x W']
        crop = self.glyph_embedding.GlyphTuple(
            *[self.crop(g, coordinates) for g in glyphs]
        )
        # -- [B x H' x W' x K]
        crop_emb = self.glyph_embedding(crop)

        if self.crop_model == "transformer":
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb, mask=None)
        elif self.crop_model == "cnn":
            # -- [B x K x W' x H']
            crop_emb = crop_emb.transpose(1, 3)
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb)
        # -- [B x K']

        crop_rep = crop_rep.view(T * B, -1)
        if self.equalize_input_dim:
            crop_rep = self.project_crop_dim(crop_rep)
        assert crop_rep.shape[0] == T * B

        reps.append(crop_rep)

        # -- [B x H x W x K]
        glyphs_emb = self.glyph_embedding(glyphs)
        # glyphs_emb = self.embed(glyphs)
        # -- [B x K x W x H]
        glyphs_emb = glyphs_emb.transpose(1, 3)
        # -- [B x W x H x K]
        glyphs_rep = self.extract_representation(glyphs_emb)

        # -- [B x K']
        glyphs_rep = glyphs_rep.view(T * B, -1)
        if self.equalize_input_dim:
            glyphs_rep = self.project_glyph_dim(glyphs_rep)

        assert glyphs_rep.shape[0] == T * B

        # -- [B x K'']
        reps.append(glyphs_rep)

        # MESSAGING MODEL
        if self.msg_model != "none":
            # [T x B x 256] -> [T * B x 256]
            messages = inputs["message"].long().view(T * B, -1)
            if self.msg_model == "cnn":
                # convert messages to one-hot, [T * B x 96 x 256]
                one_hot = F.one_hot(messages, num_classes=NUM_CHARS).transpose(
                    1, 2
                )
                char_rep = self.conv2_6_fc(self.conv1(one_hot.float()))
            elif self.msg_model == "lt_cnn":
                # [ T * B x E x 256 ]
                char_emb = self.char_lt(messages).transpose(1, 2)
                char_rep = self.conv2_6_fc(self.conv1(char_emb))
            else:  # lstm, gru
                char_emb = self.char_lt(messages)
                output = self.char_rnn(char_emb)[0]
                fwd_rep = output[:, -1, : self.h_dim // 2]
                bwd_rep = output[:, 0, self.h_dim // 2 :]
                char_rep = torch.cat([fwd_rep, bwd_rep], dim=1)

            if self.equalize_input_dim:
                char_rep = self.project_msg_dim(char_rep)
            reps.append(char_rep)

        st = torch.cat(reps, dim=1)

        # -- [B x K]
        st = self.fc(st)

        if self.use_lstm:
            core_input = st.view(T, B, -1)
            core_output_list = []
            notdone = (~inputs["done"]).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = tuple(nd * t for t in core_state)
                output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = st

        # -- [B x A]
        policy_logits = self.policy(core_output)
        # -- [B x A]
        baseline = self.baseline(core_output)

        if self.training:
            action = torch.multinomial(
                F.softmax(policy_logits, dim=1), num_samples=1
            )
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        output = dict(
            policy_logits=policy_logits, baseline=baseline, action=action
        )
        return (output, core_state)



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackSkill
from minihack.envs import register


class MiniHackQuestEasy(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 500)
        kwargs["autopickup"] = kwargs.pop("autopickup", True)
        super().__init__(*args, des_file="quest_easy.des", **kwargs)


class MiniHackQuestMedium(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 1000)
        kwargs["character"] = "kni-hum-law-fem"  # tested on human knight
        kwargs["autopickup"] = kwargs.pop("autopickup", True)
        super().__init__(*args, des_file="quest_medium.des", **kwargs)


class MiniHackQuestHard(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 1000)
        super().__init__(*args, des_file="quest_hard.des", **kwargs)


register(
    id="MiniHack-Quest-Easy-v0",
    entry_point="minihack.envs.skills_quest:MiniHackQuestEasy",
)
register(
    id="MiniHack-Quest-Medium-v0",
    entry_point="minihack.envs.skills_quest:MiniHackQuestMedium",
)
register(
    id="MiniHack-Quest-Hard-v0",
    entry_point="minihack.envs.skills_quest:MiniHackQuestHard",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Run with OMP_NUM_THREADS=1.
#

import argparse
import collections
import logging
import omegaconf
import os
import threading
import time
import timeit
import traceback
import wandb

import nest
import torch
from minihack.agent.polybeast.core import file_writer
from minihack.agent.polybeast.core import vtrace
from minihack.agent.polybeast.models import create_model, losses
from minihack.agent.polybeast.models.base import NetHackNet
from minihack.agent.polybeast.models.intrinsic import IntrinsicRewardNet
import libtorchbeast
from torch import nn
from torch.nn import functional as F

# yapf: disable
parser = argparse.ArgumentParser(description="PyTorch Scalable Agent")

parser.add_argument("--mode", default="train",
                    choices=["train", "test", "test_render"],
                    help="Training or test mode.")
parser.add_argument('--env', default='staircase', type=str, metavar='E',
                    help='Name of Gym environment to create.')
parser.add_argument("--wandb", action="store_true",
                    help="Log to wandb.")
parser.add_argument('--group', default='default', type=str, metavar='G',
                    help='Name of the experiment group (as being used by wandb).')
parser.add_argument('--project', default='minihack', type=str, metavar='P',
                    help='Name of the project (as being used by wandb).')
parser.add_argument('--entity', default='nethack', type=str, metavar='P',
                    help='Which team to log to.')

# Training settings.
parser.add_argument("--pipes_basename", default="unix:/tmp/polybeast",
                    help="Basename for the pipes for inter-process communication. "
                    "Has to be of the type unix:/some/path.")
parser.add_argument("--savedir", default="~/palaas/torchbeast",
                    help="Root dir where experiment data will be saved.")
parser.add_argument("--num_actors", default=4, type=int, metavar="N",
                    help="Number of actors")
parser.add_argument("--total_steps", default=1e6, type=float, metavar="T",
                    help="Total environment steps to train for. Will be cast to int.")
parser.add_argument("--batch_size", default=8, type=int, metavar="B",
                    help="Learner batch size")
parser.add_argument("--unroll_length", default=80, type=int, metavar="T",
                    help="The unroll length (time dimension)")
parser.add_argument("--num_learner_threads", default=2, type=int,
                    metavar="N", help="Number learner threads.")
parser.add_argument("--num_inference_threads", default=2, type=int,
                    metavar="N", help="Number learner threads.")
parser.add_argument("--learner_device", default="cuda:0", help="Set learner device")
parser.add_argument("--actor_device", default="cuda:1", help="Set actor device")
parser.add_argument("--disable_cuda", action="store_true",
                    help="Disable CUDA.")
parser.add_argument("--use_lstm", action="store_true",
                    help="Use LSTM in agent model.")
parser.add_argument("--use_index_select", action="store_true",
                    help="Whether to use index_select instead of embedding lookup.")
parser.add_argument("--max_learner_queue_size", default=None, type=int, metavar="N",
                    help="Optional maximum learner queue size. "
                    "Defaults to batch_size.")


# Model settings.
parser.add_argument('--model', default="baseline",
                    help='Name of the model to run')
parser.add_argument('--crop_model', default="cnn", choices=["cnn", "transformer"],
                    help='Size of cropping window around the agent')
parser.add_argument('--crop_dim', type=int, default=9,
                    help='Size of cropping window around the agent')
parser.add_argument('--embedding_dim', type=int, default=32,
                    help='Size of glyph embeddings.')
parser.add_argument('--hidden_dim', type=int, default=128,
                    help='Size of hidden representations.')
parser.add_argument('--layers', type=int, default=5,
                    help='Number of ConvNet/Transformer layers.')
# Loss settings.
parser.add_argument("--entropy_cost", default=0.0006, type=float,
                    help="Entropy cost/multiplier.")
parser.add_argument("--baseline_cost", default=0.5, type=float,
                    help="Baseline cost/multiplier.")
parser.add_argument("--discounting", default=0.99, type=float,
                    help="Discounting factor.")
parser.add_argument("--reward_clipping", default="tim",
                    choices=["soft_asymmetric", "none", "tim"],
                    help="Reward clipping.")
parser.add_argument("--no_extrinsic", action="store_true",
                    help=("Disables extrinsic reward (no baseline/pg_loss)."))
parser.add_argument("--normalize_reward", action="store_true",
                    help=("Normalizes reward by dividing by running stdev from "
                          "mean."))

# Optimizer settings.
parser.add_argument("--learning_rate", default=0.00048, type=float,
                    metavar="LR", help="Learning rate.")
parser.add_argument("--alpha", default=0.99, type=float,
                    help="RMSProp smoothing constant.")
parser.add_argument("--momentum", default=0, type=float,
                    help="RMSProp momentum.")
parser.add_argument("--epsilon", default=0.01, type=float,
                    help="RMSProp epsilon.")
parser.add_argument("--grad_norm_clipping", default=40.0, type=float,
                    help="Global gradient norm clip.")

# Misc settings.
parser.add_argument("--write_profiler_trace", action="store_true",
                    help="Collect and write a profiler trace "
                    "for chrome://tracing/.")

# yapf: enable


logging.basicConfig(
    format=(
        "[%(levelname)s:%(process)d %(module)s:%(lineno)d %(asctime)s] "
        "%(message)s"
    ),
    level=0,
)


def inference(
    inference_batcher, model, flags, actor_device, lock=threading.Lock()
):  # noqa: B008
    with torch.no_grad():
        for batch in inference_batcher:
            batched_env_outputs, agent_state = batch.get_inputs()
            observation, reward, done, *_ = batched_env_outputs
            # Observation is a dict with keys 'features' and 'glyphs'.
            observation["done"] = done
            observation, agent_state = nest.map(
                lambda t: t.to(actor_device, non_blocking=True),
                (observation, agent_state),
            )
            with lock:
                outputs = model(observation, agent_state)
            core_outputs, agent_state = nest.map(lambda t: t.cpu(), outputs)
            # Restructuring the output in the way that is expected
            # by the functions in actorpool.
            outputs = (
                tuple(
                    (
                        core_outputs["action"],
                        core_outputs["policy_logits"],
                        core_outputs["baseline"],
                    )
                ),
                agent_state,
            )
            batch.set_outputs(outputs)


EnvOutput = collections.namedtuple(
    "EnvOutput", "frame rewards done episode_step episode_return"
)
AgentOutput = NetHackNet.AgentOutput
Batch = collections.namedtuple("Batch", "env agent")


def clip(flags, rewards):
    if flags.reward_clipping == "tim":
        clipped_rewards = torch.tanh(rewards / 100.0)
    elif flags.reward_clipping == "soft_asymmetric":
        squeezed = torch.tanh(rewards / 5.0)
        # Negative rewards are given less weight than positive rewards.
        clipped_rewards = (
            torch.where(rewards < 0, 0.3 * squeezed, squeezed) * 5.0
        )
    elif flags.reward_clipping == "none":
        clipped_rewards = rewards
    else:
        raise NotImplementedError("reward_clipping=%s" % flags.reward_clipping)
    return clipped_rewards


def learn(
    learner_queue,
    model,
    actor_model,
    optimizer,
    scheduler,
    stats,
    flags,
    plogger,
    learner_device,
    lock=threading.Lock(),  # noqa: B008
):
    for tensors in learner_queue:
        tensors = nest.map(lambda t: t.to(learner_device), tensors)

        batch, initial_agent_state = tensors
        env_outputs, actor_outputs = batch
        observation, reward, done, *_ = env_outputs
        observation["reward"] = reward
        observation["done"] = done

        lock.acquire()  # Only one thread learning at a time.

        output, _ = model(observation, initial_agent_state, learning=True)

        # Use last baseline value (from the value function) to bootstrap.
        learner_outputs = AgentOutput._make(
            (output["action"], output["policy_logits"], output["baseline"])
        )

        # At this point, the environment outputs at time step `t` are the inputs
        # that lead to the learner_outputs at time step `t`. After the following
        # shifting, the actions in `batch` and `learner_outputs` at time
        # step `t` is what leads to the environment outputs at time step `t`.
        batch = nest.map(lambda t: t[1:], batch)
        learner_outputs = nest.map(lambda t: t[:-1], learner_outputs)

        # Turn into namedtuples again.
        env_outputs, actor_outputs = batch
        # Note that the env_outputs.frame is now a dict with 'features' and 'glyphs'
        # instead of actually being the frame itself. This is currently not a problem
        # because we never use actor_outputs.frame in the rest of this function.
        env_outputs = EnvOutput._make(env_outputs)
        actor_outputs = AgentOutput._make(actor_outputs)
        learner_outputs = AgentOutput._make(learner_outputs)

        rewards = env_outputs.rewards
        if flags.normalize_reward:
            model.update_running_moments(rewards)
            rewards /= model.get_running_std()

        total_loss = 0

        # INTRINSIC REWARDS
        calculate_intrinsic = (
            isinstance(model, IntrinsicRewardNet) and model.intrinsic_enabled()
        )
        if calculate_intrinsic:
            # Compute intrinsic reward and loss
            if "int_baseline" not in output:
                raise RuntimeError("Expected intrinsic outputs but found none")

            # set intrinsic reward dimensions here so we don't make any mistakes later
            intrinsic_reward = rewards.new_zeros(rewards.size()).float()

            if flags.model == "rnd":
                # Random Network Distillation
                target = output["target"][1:]
                predicted = output["predicted"][1:]
                # loss for prediction failures, not really "forward" model
                forward_loss = flags.rnd.forward_cost * F.mse_loss(
                    target, predicted, reduction="mean"
                )
                total_loss += forward_loss

                # reward based on unpredicted scenarios
                intrinsic_reward += (target - predicted).pow(2).sum(2) * 0.5
            elif flags.model == "ride":
                # Rewarding Impact-Driven Exploration
                state_emb = output["state_embedding"][:-1]
                next_state_emb = output["state_embedding"][1:]
                actions = actor_outputs.action

                pred_next_state_emb = model.forward_dynamics_model(
                    state_emb, actions
                )
                pred_actions = model.inverse_dynamics_model(
                    state_emb, next_state_emb
                )

                forward_loss = (
                    flags.ride.forward_cost
                    * losses.compute_forward_dynamics_loss(
                        pred_next_state_emb, next_state_emb
                    )
                )
                inverse_loss = (
                    flags.ride.inverse_cost
                    * losses.compute_inverse_dynamics_loss(
                        pred_actions, actions
                    )
                )
                total_loss += forward_loss + inverse_loss

                intrinsic_reward += torch.norm(
                    next_state_emb - state_emb, dim=2, p=2
                )
                if flags.ride.count_norm:
                    if "state_visits" not in observation:
                        raise RuntimeError(
                            "ride.count_norm=true but state_counter=none"
                        )
                    # -- [T x B ]
                    counts = (
                        observation["state_visits"][1:]
                        .squeeze(-1)
                        .float()
                        .sqrt()
                    )
                    intrinsic_reward /= counts

            if flags.int.normalize_reward:
                model.update_intrinsic_moments(intrinsic_reward)
                intrinsic_reward /= model.get_intrinsic_std()
            intrinsic_reward *= flags.int.intrinsic_weight

            if not flags.int.twoheaded and not flags.no_extrinsic:
                # add intrinsic rewards to extrinsic ones
                rewards += intrinsic_reward

        # STANDARD EXTRINSIC LOSSES / REWARDS
        if flags.entropy_cost > 0:
            entropy_loss = flags.entropy_cost * losses.compute_entropy_loss(
                learner_outputs.policy_logits
            )
            total_loss += entropy_loss

        if not flags.no_extrinsic:
            clipped_rewards = clip(flags, rewards)

            discounts = (~env_outputs.done).float() * flags.discounting

            # This could be in C++. In TF, this is actually slower on the GPU.
            vtrace_returns = vtrace.from_logits(
                behavior_policy_logits=actor_outputs.policy_logits,
                target_policy_logits=learner_outputs.policy_logits,
                actions=actor_outputs.action,
                discounts=discounts,
                rewards=clipped_rewards,
                values=learner_outputs.baseline,
                bootstrap_value=learner_outputs.baseline[-1],
            )

            # Compute loss as a weighted sum of the baseline loss, the policy
            # gradient loss and an entropy regularization term.
            pg_loss = losses.compute_policy_gradient_loss(
                learner_outputs.policy_logits,
                actor_outputs.action,
                vtrace_returns.pg_advantages,
            )
            baseline_loss = flags.baseline_cost * losses.compute_baseline_loss(
                vtrace_returns.vs - learner_outputs.baseline
            )
            total_loss += pg_loss + baseline_loss

        # TWO-HEADED INTRINSIC REWARDS / LOSSES
        if calculate_intrinsic and (flags.int.twoheaded or flags.no_extrinsic):
            # here we calculate RL loss on the int reward using its own value head
            # 1) twoheaded always separates ext and int rewards to their own heads
            # 2) no_extrinsic skips the ext value head and uses only the int one
            int_clipped_rewards = clip(flags, intrinsic_reward)

            # use a separate discounting factor for intrinsic rewards
            if flags.int.episodic:
                int_discounts = (
                    ~env_outputs.done
                ).float() * flags.int.discounting
            else:
                # can also do non-episodic intrinsic rewards
                int_discounts = discounts.new_full(
                    discounts.size(), flags.int.discounting
                )

            int_vtrace_returns = vtrace.from_logits(
                behavior_policy_logits=actor_outputs.policy_logits,
                target_policy_logits=learner_outputs.policy_logits,
                actions=actor_outputs.action,
                discounts=int_discounts,  # intrinsic discounts
                rewards=int_clipped_rewards,  # intrinsic reward
                values=output["int_baseline"][1:],  # intrinsic baseline
                bootstrap_value=output["int_baseline"][
                    -1
                ],  # intrinsic bootstrap
            )

            # intrinsic baseline loss
            int_baseline_loss = (
                flags.int.baseline_cost
                * losses.compute_baseline_loss(
                    int_vtrace_returns.vs - output["int_baseline"][1:]
                )
            )

            # intrinsic policy gradient loss
            int_pg_loss = losses.compute_policy_gradient_loss(
                learner_outputs.policy_logits,
                actor_outputs.action,
                int_vtrace_returns.pg_advantages,
            )

            total_loss += int_pg_loss + int_baseline_loss

        # BACKWARD STEP
        optimizer.zero_grad()
        total_loss.backward()
        if flags.grad_norm_clipping > 0:
            nn.utils.clip_grad_norm_(
                model.parameters(), flags.grad_norm_clipping
            )
        optimizer.step()
        scheduler.step()

        actor_model.load_state_dict(model.state_dict())

        # LOGGING
        episode_returns = env_outputs.episode_return[env_outputs.done]
        stats["step"] = (
            stats.get("step", 0) + flags.unroll_length * flags.batch_size
        )
        stats["mean_episode_return"] = torch.mean(episode_returns).item()
        stats["mean_episode_step"] = torch.mean(
            env_outputs.episode_step.float()
        ).item()
        stats["total_loss"] = total_loss.item()
        if flags.entropy_cost > 0:
            stats["entropy_loss"] = entropy_loss.item()
        if not flags.no_extrinsic:
            stats["pg_loss"] = pg_loss.item()
            stats["baseline_loss"] = baseline_loss.item()

        stats["learner_queue_size"] = learner_queue.size()

        if calculate_intrinsic:
            stats["intrinsic_reward"] = torch.mean(intrinsic_reward).item()
            if flags.model == "rnd":
                stats["forward_loss"] = forward_loss.item()
            elif flags.model == "ride":
                stats["forward_loss"] = forward_loss.item()
                stats["inverse_loss"] = inverse_loss.item()
            if flags.int.twoheaded:
                stats["int_baseline_loss"] = int_baseline_loss.item()
                stats["int_pg_loss"] = int_pg_loss.item()

        if "state_visits" in observation:
            visits = observation["state_visits"][:-1]
            metric = visits[env_outputs.done].float()
            key1 = "mean_state_visits"
            key2 = "max_state_visits"
            if not len(episode_returns):
                stats[key1] = None
                stats[key2] = None
            else:
                stats[key1] = torch.mean(metric).item()
                stats[key2] = torch.max(metric).item()

        DEBUG = False

        if DEBUG and env_outputs.done.sum() > 0:
            print()
            print("glyphs shape", env_outputs.frame["glyphs"].shape)
            print("features shape", env_outputs.frame["features"].shape)
            print(
                "episode_step",
                env_outputs.episode_step[:, 0],
                env_outputs.episode_step.shape,
            )
            print(
                "rewards", env_outputs.rewards[:, 0], env_outputs.rewards.shape
            )
            print(
                "episode_return",
                env_outputs.episode_return[:, 0],
                env_outputs.episode_return.shape,
            )
            print("done", env_outputs.done[:, 0], env_outputs.done.shape)

        if not len(episode_returns):
            # Hide the mean-of-empty-tuple NaN as it scares people.
            stats["mean_episode_return"] = None

        # Only logging if at least one episode was finished
        if len(episode_returns):
            plogger.log(stats)
            if flags.wandb:
                wandb.log(stats, step=stats["step"])

        lock.release()


def train(flags):
    logging.info("Logging results to %s", flags.savedir)
    if isinstance(flags, omegaconf.DictConfig):
        flag_dict = omegaconf.OmegaConf.to_container(flags)
    else:
        flag_dict = vars(flags)
    plogger = file_writer.FileWriter(xp_args=flag_dict, rootdir=flags.savedir)

    if not flags.disable_cuda and torch.cuda.is_available():
        logging.info("Using CUDA.")
        learner_device = torch.device(flags.learner_device)
        actor_device = torch.device(flags.actor_device)
    else:
        logging.info("Not using CUDA.")
        learner_device = torch.device("cpu")
        actor_device = torch.device("cpu")

    if flags.max_learner_queue_size is None:
        flags.max_learner_queue_size = flags.batch_size

    # The queue the learner threads will get their data from.
    # Setting `minimum_batch_size == maximum_batch_size`
    # makes the batch size static. We could make it dynamic, but that
    # requires a loss (and learning rate schedule) that's batch size
    # independent.
    learner_queue = libtorchbeast.BatchingQueue(
        batch_dim=1,
        minimum_batch_size=flags.batch_size,
        maximum_batch_size=flags.batch_size,
        check_inputs=True,
        maximum_queue_size=flags.max_learner_queue_size,
    )

    # The "batcher", a queue for the inference call. Will yield
    # "batch" objects with `get_inputs` and `set_outputs` methods.
    # The batch size of the tensors will be dynamic.
    inference_batcher = libtorchbeast.DynamicBatcher(
        batch_dim=1,
        minimum_batch_size=1,
        maximum_batch_size=512,
        timeout_ms=100,
        check_outputs=True,
    )

    addresses = []
    connections_per_server = 1
    pipe_id = 0
    while len(addresses) < flags.num_actors:
        for _ in range(connections_per_server):
            addresses.append(f"{flags.pipes_basename}.{pipe_id}")
            if len(addresses) == flags.num_actors:
                break
        pipe_id += 1

    logging.info("Using model %s", flags.model)

    model = create_model(flags, learner_device)

    plogger.metadata["model_numel"] = sum(
        p.numel() for p in model.parameters() if p.requires_grad
    )

    logging.info(
        "Number of model parameters: %i", plogger.metadata["model_numel"]
    )

    actor_model = create_model(flags, actor_device)

    # The ActorPool that will run `flags.num_actors` many loops.
    actors = libtorchbeast.ActorPool(
        unroll_length=flags.unroll_length,
        learner_queue=learner_queue,
        inference_batcher=inference_batcher,
        env_server_addresses=addresses,
        initial_agent_state=model.initial_state(),
    )

    def run():
        try:
            actors.run()
        except Exception as e:
            logging.error("Exception in actorpool thread!")
            traceback.print_exc()
            print()
            raise e

    actorpool_thread = threading.Thread(target=run, name="actorpool-thread")

    optimizer = torch.optim.RMSprop(
        model.parameters(),
        lr=flags.learning_rate,
        momentum=flags.momentum,
        eps=flags.epsilon,
        alpha=flags.alpha,
    )

    def lr_lambda(epoch):
        return (
            1
            - min(
                epoch * flags.unroll_length * flags.batch_size,
                flags.total_steps,
            )
            / flags.total_steps
        )

    scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)

    stats = {}

    if flags.checkpoint and os.path.exists(flags.checkpoint):
        logging.info("Loading checkpoint: %s" % flags.checkpoint)
        checkpoint_states = torch.load(
            flags.checkpoint, map_location=flags.learner_device
        )
        model.load_state_dict(checkpoint_states["model_state_dict"])
        optimizer.load_state_dict(checkpoint_states["optimizer_state_dict"])
        scheduler.load_state_dict(checkpoint_states["scheduler_state_dict"])
        stats = checkpoint_states["stats"]
        logging.info(f"Resuming preempted job, current stats:\n{stats}")

    # Initialize actor model like learner model.
    actor_model.load_state_dict(model.state_dict())

    learner_threads = [
        threading.Thread(
            target=learn,
            name="learner-thread-%i" % i,
            args=(
                learner_queue,
                model,
                actor_model,
                optimizer,
                scheduler,
                stats,
                flags,
                plogger,
                learner_device,
            ),
        )
        for i in range(flags.num_learner_threads)
    ]
    inference_threads = [
        threading.Thread(
            target=inference,
            name="inference-thread-%i" % i,
            args=(inference_batcher, actor_model, flags, actor_device),
        )
        for i in range(flags.num_inference_threads)
    ]

    actorpool_thread.start()
    for t in learner_threads + inference_threads:
        t.start()

    def checkpoint(checkpoint_path=None):
        if flags.checkpoint:
            if checkpoint_path is None:
                checkpoint_path = flags.checkpoint
            logging.info("Saving checkpoint to %s", checkpoint_path)
            torch.save(
                {
                    "model_state_dict": model.state_dict(),
                    "optimizer_state_dict": optimizer.state_dict(),
                    "scheduler_state_dict": scheduler.state_dict(),
                    "stats": stats,
                    "flags": vars(flags),
                },
                checkpoint_path,
            )

    def format_value(x):
        return f"{x:1.5}" if isinstance(x, float) else str(x)

    try:
        train_start_time = timeit.default_timer()
        train_time_offset = stats.get(
            "train_seconds", 0
        )  # used for resuming training
        last_checkpoint_time = timeit.default_timer()

        dev_checkpoint_intervals = [0, 0.25, 0.5, 0.75]

        loop_start_time = timeit.default_timer()
        loop_start_step = stats.get("step", 0)
        while True:
            if loop_start_step >= flags.total_steps:
                break
            time.sleep(5)
            loop_end_time = timeit.default_timer()
            loop_end_step = stats.get("step", 0)

            stats["train_seconds"] = round(
                loop_end_time - train_start_time + train_time_offset, 1
            )

            if loop_end_time - last_checkpoint_time > 10 * 60:
                # Save every 10 min.
                checkpoint()
                last_checkpoint_time = loop_end_time

            if len(dev_checkpoint_intervals) > 0:
                step_percentage = loop_end_step / flags.total_steps
                i = dev_checkpoint_intervals[0]
                if step_percentage > i:
                    checkpoint(flags.checkpoint[:-4] + "_" + str(i) + ".tar")
                    dev_checkpoint_intervals = dev_checkpoint_intervals[1:]

            logging.info(
                "Step %i @ %.1f SPS. Inference batcher size: %i."
                " Learner queue size: %i."
                " Other stats: (%s)",
                loop_end_step,
                (loop_end_step - loop_start_step)
                / (loop_end_time - loop_start_time),
                inference_batcher.size(),
                learner_queue.size(),
                ", ".join(
                    f"{key} = {format_value(value)}"
                    for key, value in stats.items()
                ),
            )
            loop_start_time = loop_end_time
            loop_start_step = loop_end_step
    except KeyboardInterrupt:
        pass  # Close properly.
    else:
        logging.info("Learning finished after %i steps.", stats["step"])

    checkpoint()

    # Done with learning. Let's stop all the ongoing work.
    inference_batcher.close()
    learner_queue.close()

    actorpool_thread.join()

    for t in learner_threads + inference_threads:
        t.join()


def test(flags):
    test_checkpoint = os.path.join(flags.savedir, "test_checkpoint.tar")

    if not os.path.exists(os.path.dirname(test_checkpoint)):
        os.makedirs(os.path.dirname(test_checkpoint))

    logging.info("Creating test copy of checkpoint '%s'", flags.checkpoint)

    checkpoint = torch.load(flags.checkpoint)
    for d in checkpoint["optimizer_state_dict"]["param_groups"]:
        d["lr"] = 0.0
        d["initial_lr"] = 0.0

    checkpoint["scheduler_state_dict"]["last_epoch"] = 0
    checkpoint["scheduler_state_dict"]["_step_count"] = 0
    checkpoint["scheduler_state_dict"]["base_lrs"] = [0.0]
    checkpoint["stats"]["step"] = 0
    checkpoint["stats"]["_tick"] = 0

    flags.checkpoint = test_checkpoint
    flags.learning_rate = 0.0

    logging.info("Saving test checkpoint to %s", test_checkpoint)
    torch.save(checkpoint, test_checkpoint)

    train(flags)


def main(flags):
    if flags.wandb:
        flags_dict = omegaconf.OmegaConf.to_container(flags)
        wandb.init(
            project=flags.project,
            config=flags_dict,
            group=flags.group,
            entity=flags.entity,
        )
    if flags.mode == "train":
        if flags.write_profiler_trace:
            logging.info("Running with profiler.")
            with torch.autograd.profiler.profile() as prof:
                train(flags)
            filename = "chrome-%s.trace" % time.strftime("%Y%m%d-%H%M%S")
            logging.info("Writing profiler trace to '%s.gz'", filename)
            prof.export_chrome_trace(filename)
            os.system("gzip %s" % filename)
        else:
            train(flags)
    elif flags.mode.startswith("test"):
        test(flags)


if __name__ == "__main__":
    flags = parser.parse_args()
    flags.total_steps = int(flags.total_steps)  # Allows e.g. 1e6.
    main(flags)



# Copyright (c) Facebook, Inc. and its affiliates.

import threading
from nle.env import tasks as nle_tasks
from minihack import MiniHack
from minihack.envs import (
    corridor,
    keyroom,
    mazewalk,
    minigrid,
    room,
    boxohack,
    fightcorridor,
    river,
    memento,
    hidenseek,
    exploremaze,
    skills_simple,
    skills_lava,
    skills_wod,
    skills_quest,
)
from minihack.agent.common.envs.wrapper import (
    CounterWrapper,
    CropWrapper,
    PrevWrapper,
)


ENVS = dict(
    # NLE tasks
    staircase=nle_tasks.NetHackStaircase,
    score=nle_tasks.NetHackScore,
    pet=nle_tasks.NetHackStaircasePet,
    oracle=nle_tasks.NetHackOracle,
    gold=nle_tasks.NetHackGold,
    eat=nle_tasks.NetHackEat,
    scout=nle_tasks.NetHackScout,
    # MiniHack Room
    small_room=room.MiniHackRoom5x5,
    small_room_random=room.MiniHackRoom5x5Random,
    small_room_dark=room.MiniHackRoom5x5Dark,
    small_room_monster=room.MiniHackRoom5x5Monster,
    small_room_trap=room.MiniHackRoom5x5Trap,
    small_room_ultimate=room.MiniHackRoom5x5Ultimate,
    big_room=room.MiniHackRoom15x15,
    big_room_random=room.MiniHackRoom15x15Random,
    big_room_dark=room.MiniHackRoom15x15Dark,
    big_room_monster=room.MiniHackRoom15x15Monster,
    big_room_trap=room.MiniHackRoom15x15Trap,
    big_room_ultimate=room.MiniHackRoom15x15Ultimate,
    # MiniHack Corridor
    corridor2=corridor.MiniHackCorridor2,
    corridor3=corridor.MiniHackCorridor3,
    corridor5=corridor.MiniHackCorridor5,
    # MiniHack KeyRoom
    keyroom_small_fixed=keyroom.MiniHackKeyRoom5x5Fixed,
    keyroom_small=keyroom.MiniHackKeyRoom5x5,
    keyroom_small_dark=keyroom.MiniHackKeyRoom5x5Dark,
    keyroom_big=keyroom.MiniHackKeyRoom15x15,
    keyroom_big_dark=keyroom.MiniHackKeyRoom15x15Dark,
    # MiniHack MazeWalk
    mazewalk_small=mazewalk.MiniHackMazeWalk9x9,
    mazewalk_small_mapped=mazewalk.MiniHackMazeWalk9x9Premapped,
    mazewalk_big=mazewalk.MiniHackMazeWalk15x15,
    mazewalk_big_mapped=mazewalk.MiniHackMazeWalk15x15Premapped,
    mazewalk_huge=mazewalk.MiniHackMazeWalk45x19,
    mazewalk_huge_mapped=mazewalk.MiniHackMazeWalk45x19Premapped,
    # MiniHack Fight Corridor
    fight_corridor=fightcorridor.MiniHackFightCorridor,
    fight_corridor_dark=fightcorridor.MiniHackFightCorridorDark,
    # MiniHack River
    river=river.MiniHackRiver,
    river_lava=river.MiniHackRiverLava,
    river_monster=river.MiniHackRiverMonster,
    river_monsterlava=river.MiniHackRiverMonsterLava,
    river_narrow=river.MiniHackRiverNarrow,
    # MiniHack Memento
    memento_short=memento.MiniHackMementoShortF2,
    memento=memento.MiniHackMementoF2,
    memento_hard=memento.MiniHackMementoF4,
    # MiniHack Hide&Seek
    hidenseek=hidenseek.MiniHackHideAndSeek,
    hidenseek_mapped=hidenseek.MiniHackHideAndSeekMapped,
    hidenseek_lava=hidenseek.MiniHackHideAndSeekLava,
    hidenseek_big=hidenseek.MiniHackHideAndSeekBig,
    # MiniHack ExploreMaze
    explore_easy=exploremaze.MiniHackExploreMazeEasy,
    explore_easy_map=exploremaze.MiniHackExploreMazeEasyMapped,
    explore_hard=exploremaze.MiniHackExploreMazeHard,
    explore_hard_map=exploremaze.MiniHackExploreMazeHardMapped,
    # MiniHack MultiRooms
    multiroom_2=minigrid.MiniHackMultiRoomN2,
    multiroom_4=minigrid.MiniHackMultiRoomN4,
    multiroom_6=minigrid.MiniHackMultiRoomN6,
    multiroom_2_locked=minigrid.MiniHackMultiRoomN2Locked,
    multiroom_4_locked=minigrid.MiniHackMultiRoomN4Locked,
    multiroom_6_locked=minigrid.MiniHackMultiRoomN6Locked,
    multiroom_2_lava=minigrid.MiniHackMultiRoomN2Lava,
    multiroom_4_lava=minigrid.MiniHackMultiRoomN4Lava,
    multiroom_6_lava=minigrid.MiniHackMultiRoomN6Lava,
    multiroom_2_monster=minigrid.MiniHackMultiRoomN2Monster,
    multiroom_4_monster=minigrid.MiniHackMultiRoomN4Monster,
    multiroom_6_monster=minigrid.MiniHackMultiRoomN6Monster,
    multiroom_2_extreme=minigrid.MiniHackMultiRoomN2Extreme,
    multiroom_4_extreme=minigrid.MiniHackMultiRoomN4Extreme,
    multiroom_6_extreme=minigrid.MiniHackMultiRoomN6Extreme,
    # MiniHack Boxoban
    boxoban_unfiltered=boxohack.MiniHackBoxobanUnfiltered,
    boxoban_hard=boxohack.MiniHackBoxobanHard,
    boxoban_medium=boxohack.MiniHackBoxobanMedium,
    # MiniHack Simple Skills
    mini_eat=skills_simple.MiniHackEat,
    mini_pray=skills_simple.MiniHackPray,
    mini_sink=skills_simple.MiniHackSink,
    mini_read=skills_simple.MiniHackRead,
    mini_zap=skills_simple.MiniHackZap,
    mini_puton=skills_simple.MiniHackPutOn,
    mini_wear=skills_simple.MiniHackWear,
    mini_wield=skills_simple.MiniHackWield,
    mini_locked=skills_simple.MiniHackLockedDoor,
    # MiniHack Simple Skills (Fixed versions)
    mini_eat_fixed=skills_simple.MiniHackEatFixed,
    mini_pray_fixed=skills_simple.MiniHackPrayFixed,
    mini_sink_fixed=skills_simple.MiniHackSinkFixed,
    mini_read_fixed=skills_simple.MiniHackReadFixed,
    mini_zap_fixed=skills_simple.MiniHackZapFixed,
    mini_puton_fixed=skills_simple.MiniHackPutOnFixed,
    mini_wear_fixed=skills_simple.MiniHackWearFixed,
    mini_wield_fixed=skills_simple.MiniHackWieldFixed,
    mini_locked_fixed=skills_simple.MiniHackLockedDoorFixed,
    # MiniHack Simple Skills (Fixed versions)
    mini_eat_distr=skills_simple.MiniHackEatDistr,
    mini_pray_distr=skills_simple.MiniHackPrayDistr,
    mini_sink_distr=skills_simple.MiniHackSinkDistr,
    mini_read_distr=skills_simple.MiniHackReadDistr,
    mini_zap_distr=skills_simple.MiniHackZapDistr,
    mini_puton_distr=skills_simple.MiniHackPutOnDistr,
    mini_wear_distr=skills_simple.MiniHackWearDistr,
    mini_wield_distr=skills_simple.MiniHackWieldDistr,
    # WoD
    wod_easy=skills_wod.MiniHackWoDEasy,
    wod_medium=skills_wod.MiniHackWoDMedium,
    wod_hard=skills_wod.MiniHackWoDHard,
    wod_pro=skills_wod.MiniHackWoDPro,
    # MiniHack Lava Crossing
    lava=skills_lava.MiniHackLC,
    lava_lev=skills_lava.MiniHackLCLevitate,
    lava_lev_potion_inv=skills_lava.MiniHackLCLevitatePotionInv,
    lava_lev_potion_pick=skills_lava.MiniHackLCLevitatePotionPickup,
    lava_lev_ring_inv=skills_lava.MiniHackLCLevitateRingInv,
    lava_lev_ring_pick=skills_lava.MiniHackLCLevitateRingPickup,
    # MiniHack Quest
    quest_easy=skills_quest.MiniHackQuestEasy,
    quest_medium=skills_quest.MiniHackQuestMedium,
    quest_hard=skills_quest.MiniHackQuestHard,
)


def is_env_minihack(env_cls):
    return issubclass(env_cls, MiniHack)


def create_env(flags, env_id=0):
    # Create environment instances for actors
    with threading.Lock():
        env_class = ENVS[flags.env]

        if flags.save_tty:
            savedir = ""  # NLE choses location
        else:
            savedir = None

        kwargs = dict(
            savedir=savedir,
            archivefile=None,
            observation_keys=flags.obs_keys.split(","),
            penalty_step=flags.penalty_step,
            penalty_time=flags.penalty_time,
            penalty_mode=flags.fn_penalty_step,
        )
        if not is_env_minihack(env_class):
            kwargs.update(max_episode_steps=flags.max_num_steps)
            kwargs.update(character=flags.character)

        env = env_class(**kwargs)
        if flags.state_counter != "none":
            env = CounterWrapper(env, flags.state_counter)
        if flags.seedspath is not None and len(flags.seedspath) > 0:
            raise NotImplementedError("seedspath > 0 not implemented yet.")

        return env



# Copyright (c) Facebook, Inc. and its affiliates.

import os
from collections.abc import Iterable
from numbers import Number

import hydra
import minihack.agent.rllib.models  # noqa: F401
from minihack.agent.common.envs import tasks
from minihack.agent import is_env_registered, get_env_shortcut
import numpy as np
import ray
import ray.tune.integration.wandb
from minihack.agent.rllib.envs import RLLibNLEEnv  # noqa: F401
from omegaconf import DictConfig, OmegaConf
from ray import tune
from ray.rllib.models.catalog import MODEL_DEFAULTS
from ray.rllib.agents import a3c, dqn, impala, ppo
from ray.tune.integration.wandb import (
    _VALID_ITERABLE_TYPES,
    _VALID_TYPES,
    WandbLoggerCallback,
)
from ray.tune.registry import register_env
from ray.tune.utils import merge_dicts


def get_full_config(cfg: DictConfig) -> DictConfig:
    env_flags = OmegaConf.to_container(cfg)
    env_flags["seedspath"] = ""
    return OmegaConf.create(env_flags)


NAME_TO_TRAINER: dict = {
    "impala": (impala.DEFAULT_CONFIG.copy(), impala.ImpalaTrainer),
    "a2c": (a3c.a2c.A2C_DEFAULT_CONFIG.copy(), a3c.A2CTrainer),
    "dqn": (dqn.DEFAULT_CONFIG.copy(), dqn.DQNTrainer),
    "ppo": (ppo.DEFAULT_CONFIG.copy(), ppo.PPOTrainer),
}


@hydra.main(config_path=".", config_name="config")
def train(cfg: DictConfig) -> None:
    ray.init(num_gpus=cfg.num_gpus, num_cpus=cfg.num_cpus + 1)
    cfg = get_full_config(cfg)
    register_env("RLlibNLE-v0", RLLibNLEEnv)

    try:
        config, trainer = NAME_TO_TRAINER[cfg.algo]
    except KeyError as error:
        raise ValueError(
            "The algorithm you specified isn't currently supported: %s",
            cfg.algo,
        ) from error

    args_config = OmegaConf.to_container(cfg)

    # Algo-specific config. Requires hydra config keys to match rllib exactly
    algo_config = args_config.pop(cfg.algo)

    # Remove unnecessary config keys
    for algo in NAME_TO_TRAINER.keys():
        if algo != cfg.algo:
            args_config.pop(algo, None)

    # Merge config from hydra (will have some rogue keys but that's ok)
    config = merge_dicts(config, args_config)

    # check the name of the environment
    if cfg.env not in tasks.ENVS:
        if is_env_registered(cfg.env):
            cfg.env = get_env_shortcut(cfg.env)
        else:
            raise KeyError(
                f"Could not find an environement with a name: {cfg.env}."
            )

    # Update configuration with parsed arguments in specific ways
    config = merge_dicts(
        config,
        {
            "framework": "torch",
            "num_gpus": cfg.num_gpus,
            "seed": cfg.seed,
            "env": "RLlibNLE-v0",
            "env_config": {
                "flags": cfg,
                "observation_keys": cfg.obs_keys.split(","),
                "name": cfg.env,
            },
            "train_batch_size": cfg.train_batch_size,
            "model": merge_dicts(
                MODEL_DEFAULTS,
                {
                    "custom_model": "rllib_nle_model",
                    "custom_model_config": {"flags": cfg, "algo": cfg.algo},
                    "use_lstm": cfg.use_lstm,
                    "lstm_use_prev_reward": True,
                    "lstm_use_prev_action": True,
                    "lstm_cell_size": cfg.hidden_dim,
                },
            ),
            "num_workers": cfg.num_cpus,
            "num_envs_per_worker": int(cfg.num_actors / cfg.num_cpus),
            "evaluation_interval": 100,
            "evaluation_num_episodes": 50,
            "evaluation_config": {"explore": False},
            "rollout_fragment_length": cfg.unroll_length,
        },
    )

    # Merge algo-specific config at top level
    config = merge_dicts(config, algo_config)

    # Ensure we can use the config we've specified above
    trainer_class = trainer.with_updates(default_config=config)

    callbacks = []
    if cfg.wandb:
        callbacks.append(
            WandbLoggerCallback(
                project=cfg.project,
                api_key_file="~/.wandb_api_key",
                entity=cfg.entity,
                group=cfg.group,
                tags=cfg.tags.split(","),
            )
        )
        os.environ[
            "TUNE_DISABLE_AUTO_CALLBACK_LOGGERS"
        ] = "1"  # Only log to wandb

    # Hacky monkey-patching to allow for OmegaConf config
    def _is_allowed_type(obj):
        """Return True if type is allowed for logging to wandb"""
        if isinstance(obj, DictConfig):
            return True
        if isinstance(obj, np.ndarray) and obj.size == 1:
            return isinstance(obj.item(), Number)
        if isinstance(obj, Iterable) and len(obj) > 0:
            return isinstance(obj[0], _VALID_ITERABLE_TYPES)
        return isinstance(obj, _VALID_TYPES)

    ray.tune.integration.wandb._is_allowed_type = _is_allowed_type

    tune.run(
        trainer_class,
        stop={"timesteps_total": cfg.total_steps},
        config=config,
        name=cfg.name,
        callbacks=callbacks,
    )


if __name__ == "__main__":
    train()



# Copyright (c) Facebook, Inc. and its affiliates.

from collections import OrderedDict
from typing import Tuple, Union

import gym
import numpy as np

from minihack.agent.common.envs.tasks import create_env


class RLLibNLEEnv(gym.Env):
    def __init__(self, env_config: dict) -> None:
        # We sort the observation keys so we can create the OrderedDict output
        # in a consistent order
        self._observation_keys = sorted(
            env_config.get("observation_keys", ("blstats", "glyphs"))
        )
        self.gym_env = create_env(env_config["flags"])

    @property
    def action_space(self) -> gym.Space:
        return self.gym_env.action_space

    @property
    def observation_space(self) -> gym.Space:
        return self.gym_env.observation_space

    def reset(self) -> dict:
        return self._process_obs(self.gym_env.reset())

    def _process_obs(self, obs: dict) -> dict:
        return OrderedDict({key: obs[key] for key in self._observation_keys})

    def step(
        self, action: Union[int, np.int64]
    ) -> Tuple[dict, Union[np.number, int], Union[np.bool_, bool], dict]:
        obs, reward, done, info = self.gym_env.step(action)
        return self._process_obs(obs), reward, done, info

    def render(self):
        return self.gym_env.render()

    def close(self):
        return self.gym_env.close()



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackSkill, LevelGenerator, RewardManager
from minihack.envs import register


class MiniHackWoDEasy(MiniHackSkill):
    """Environment for "Wand of death" task."""

    def __init__(self, *args, **kwargs):
        map = """
|----------
|.........+
|----------
"""
        lvl_gen = LevelGenerator(map=map, lit=True)

        lvl_gen.set_start_pos((1, 1))
        kwargs["autopickup"] = kwargs.pop("autopickup", True)
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 50)

        lvl_gen.add_object(
            name="death", symbol="/", cursestate="blessed", place=((1, 1))
        )
        lvl_gen.add_monster("minotaur", args=("asleep",), place=(9, 1))

        des_file = lvl_gen.get_des()

        rwrd_mngr = RewardManager()
        rwrd_mngr.add_kill_event("minotaur")

        super().__init__(
            *args, des_file=des_file, reward_manager=rwrd_mngr, **kwargs
        )


class MiniHackWoDMedium(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        map = """
|---------------------------|
|...........................|
|---------------------------|
"""
        lvl_gen = LevelGenerator(map=map, lit=True)

        lvl_gen.set_start_pos((1, 1))
        lvl_gen.add_goal_pos((27, 1))

        lvl_gen.add_object(
            name="death", symbol="/", cursestate="blessed", place=((2, 1))
        )
        lvl_gen.add_monster("minotaur", args=("asleep",), place=(26, 1))
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 150)

        des_file = lvl_gen.get_des()
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackWoDHard(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        map = """
|---------------------------|
|...........................|
|.....|---------------------|
|.....|
|.....|
|-----|
"""
        lvl_gen = LevelGenerator(map=map, lit=True)

        lvl_gen.set_start_rect((1, 1), (5, 5))
        lvl_gen.add_goal_pos((27, 1))

        lvl_gen.set_area_variable("$safe_room", "fillrect", 1, 1, 5, 5)
        lvl_gen.add_object_area(
            "$safe_room", name="death", symbol="/", cursestate="blessed"
        )
        lvl_gen.add_monster("minotaur", place=(26, 1))
        des_file = lvl_gen.get_des()
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackWoDPro(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        map = """
-------------------------------------
|.................|.|...............|
|.|-------------|.|.|.------------|.|
|.|.............|.|.|.............|.|
|.|.|----------.|.|.|------------.|.|
|.|.|...........|.|.............|.|.|
|.|.|.|----------.|-----------|.|.|.|
|.|.|.|...........|.......|...|.|.|.|
|.|.|.|.|----------------.|.|.|.|.|.|
|.|.|.|.|.................|.|.|.|.|.|
|.|.|.|.|.-----------------.|.|.|.|.|
|.|.|.|.|...................|.|.|.|.|
|.|.|.|.|--------------------.|.|.|.|
|.|.|.|.......................|.|.|.|
|.|.|.|-----------------------|.|.|.|
|.|.|...........................|.|.|
|.|.|---------------------------|.|.|
|.|...............................|.|
|.|-------------------------------|.|
|...................................|
-------------------------------------
"""
        lvl_gen = LevelGenerator(map=map, lit=True)
        lvl_gen.set_start_pos((19, 1))
        lvl_gen.add_goal_pos((19, 7))
        lvl_gen.add_monster(name="minotaur", place=(19, 9))
        lvl_gen.add_object("death", "/", cursestate="blessed")
        des_file = lvl_gen.get_des()
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 1000)

        super().__init__(
            *args,
            des_file=des_file,
            **kwargs,
        )


register(
    id="MiniHack-WoD-Easy-v0",
    entry_point="minihack.envs.skills_wod:MiniHackWoDEasy",
)

register(
    id="MiniHack-WoD-Medium-v0",
    entry_point="minihack.envs.skills_wod:MiniHackWoDMedium",
)
register(
    id="MiniHack-WoD-Hard-v0",
    entry_point="minihack.envs.skills_wod:MiniHackWoDHard",
)
register(
    id="MiniHack-WoD-Pro-v0",
    entry_point="minihack.envs.skills_wod:MiniHackWoDPro",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
import os

env_name_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "env_names.yaml"
)
with open(env_name_path, "r") as stream:
    env_short_to_full = yaml.safe_load(stream)

env_full_to_short = {v: k for k, v in env_short_to_full.items()}


def is_env_registered(env):
    return env in env_full_to_short.keys() or env in env_short_to_full.keys()


def get_env_shortcut(env_gym_name):
    try:
        return env_full_to_short[env_gym_name]
    except KeyError:
        raise KeyError(
            "Could not find an environement with a registered "
            f"name: {env_gym_name}. For the full list, see {env_name_path}."
        )


def get_env_gym_name(env_shortcut):
    try:
        return env_short_to_full[env_shortcut]
    except KeyError:
        raise KeyError(
            "Could not find an environement with a shortcut name:"
            f"{env_shortcut}. For the full list, see {env_name_path}."
        )



# Copyright (c) Facebook, Inc. and its affiliates.

import minihack.agent.polybeast.models
from minihack.agent import get_env_shortcut
from minihack.agent.polybeast import polyhydra
import argparse
import time
import timeit
import os
import tempfile
import shutil
import glob

import gym
import torch
from omegaconf import OmegaConf


def get_action(model, obs, hidden, done, watch):
    with torch.no_grad():
        for key in obs.keys():
            shape = obs[key].shape
            obs[key] = torch.Tensor(obs[key].reshape((1, 1, *shape)))

        obs["done"] = torch.BoolTensor([done])

        out, hidden = model(obs, hidden)

        action = out["action"]

    if watch:
        input()

    return action, hidden


def load_model(env, checkpoint_path):
    pretrained_path = os.path.join(checkpoint_path, "checkpoint.tar")
    pretrained_config_path = os.path.join(checkpoint_path, "config.yaml")
    flags = OmegaConf.load(pretrained_config_path)
    flags["env"] = env
    flags = polyhydra.get_common_flags(flags)
    flags = polyhydra.get_environment_flags(flags)
    flags = polyhydra.get_learner_flags(flags)
    model = minihack.agent.polybeast.models.create_model(
        flags, torch.device("cpu")
    )

    checkpoint_states = torch.load(
        pretrained_path, map_location=torch.device("cpu")
    )

    model.load_state_dict(checkpoint_states["model_state_dict"])

    hidden = model.initial_state(batch_size=1)
    model.eval()
    return model, hidden


def eval(
    env,
    num_episodes,
    max_steps,
    seeds,
    savedir,
    no_render,
    render_mode,
    checkpoint_dir,
    watch,
    obs_keys,
    save_gif,
    gif_path,
    gif_duration,
):
    agent_env = get_env_shortcut(env)
    env = gym.make(
        env,
        savedir=savedir,
        max_episode_steps=max_steps,
        observation_keys=obs_keys.split(","),
    )
    if seeds is not None:
        env.seed(seeds)
    if not no_render:
        print("Available actions:", env._actions)

    obs = env.reset()
    done = False

    model, hidden = load_model(agent_env, checkpoint_dir)

    steps = 0
    episodes = 0
    reward = 0.0
    action = None

    mean_sps = 0
    mean_reward = 0.0
    mean_return = 0.0

    total_start_time = timeit.default_timer()
    start_time = total_start_time

    if save_gif:
        # Import pillow
        try:
            import PIL.Image
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To safe GIF files of trajectories, please install Pillow:"
                " pip install Pillow"
            )
        # Create a tmp directory for individual screenshots
        tmpdir = tempfile.mkdtemp()

    while True:
        if watch and not no_render:
            print("Previous reward:", reward)
            if action is not None:
                print("Previous action: %s" % repr(env._actions[action]))
            env.render(render_mode)

        if save_gif:
            obs_image = PIL.Image.fromarray(obs["pixel_crop"])
            obs_image.save(os.path.join(tmpdir, f"e_{episodes}_s_{steps}.png"))

        action, hidden = get_action(model, obs, hidden, done, watch)
        if action is None:
            break

        obs, reward, done, info = env.step(action)
        steps += 1

        mean_reward += (reward - mean_reward) / steps

        if not done:
            continue

        time_delta = timeit.default_timer() - start_time

        print("Final reward:", reward)
        print("End status:", info["end_status"].name)
        print("Mean reward:", mean_reward)
        print("Total reward:", mean_reward * steps)

        mean_return += (mean_reward * steps) / num_episodes
        sps = steps / time_delta
        print("Episode: %i. Steps: %i. SPS: %f" % (episodes, steps, sps))

        episodes += 1
        mean_sps += (sps - mean_sps) / episodes

        start_time = timeit.default_timer()

        steps = 0
        mean_reward = 0.0

        if episodes == num_episodes:
            break
        env.reset()

    if save_gif:
        # Make the GIF and delete the temporary directory
        png_files = glob.glob(os.path.join(tmpdir, "e_*_s_*.png"))
        png_files.sort(key=os.path.getmtime)

        img, *imgs = [PIL.Image.open(f) for f in png_files]
        img.save(
            fp=gif_path,
            format="GIF",
            append_images=imgs,
            save_all=True,
            duration=gif_duration,
            loop=0,
        )
        shutil.rmtree(tmpdir)

        print("Saving replay GIF at {}".format(os.path.abspath(gif_path)))

    env.close()
    print(
        "Finished after %i episodes and %f seconds. Mean sps: %f"
        % (episodes, timeit.default_timer() - total_start_time, mean_sps)
    )
    print(f"Mean return is {mean_return}")


def main():
    parser = argparse.ArgumentParser(
        description="Tool for evaluating pretrained models."
    )
    parser.add_argument(
        "-e",
        "--env",
        type=str,
        default="MiniHack-Room-15x15-v0",
        help="Gym environment spec. Defaults to 'MiniHack-Room-15x15-v0'.",
    )
    parser.add_argument(
        "-c",
        "--checkpoint_dir",
        type=str,
        help="Path to checkpoint directory to load the model and configs. "
        + "This directory must include checkpoint.tar and config.yaml files.",
    )
    parser.add_argument(
        "-o",
        "--obs_keys",
        type=str,
        default="glyphs,chars,colors,specials,blstats,message",
        help="The observation keys as a string. Separate keys using a comma",
    )
    parser.add_argument(
        "-n",
        "--num_episodes",
        type=int,
        default=5,
        help="Number of episodes to be evaluated before exiting.",
    )
    parser.add_argument(
        "--max-steps",
        type=int,
        default=10000,
        help="Number of maximum steps per episode.",
    )
    parser.add_argument(
        "--seeds",
        default=None,
        help="Seeds to send to NetHack. Can be a dict or int. "
        "Defaults to None (no seeding).",
    )
    parser.add_argument(
        "--savedir",
        default=None,
        help="Directory path where data will be saved. "
        "Defaults to None (not data saved).",
    )
    parser.add_argument(
        "--no-render", action="store_true", help="Disables env.render()."
    )
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "full", "ansi"],
        help="Render mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "--watch",
        dest="watch",
        action="store_true",
        help="Pressing the Enter key performs a step.",
    )
    parser.add_argument(
        "--no-watch",
        dest="watch",
        action="store_false",
        help="Not watching the replay.",
    )
    parser.set_defaults(watch=True)
    parser.add_argument(
        "--save_gif",
        dest="save_gif",
        action="store_true",
        help="Saving a GIF replay of the evaluated episodes.",
    )
    parser.add_argument(
        "--no-save_gif",
        dest="save_gif",
        action="store_false",
        help="Do not save GIF.",
    )
    parser.set_defaults(save_gif=False)
    parser.add_argument(
        "--gif_path",
        type=str,
        default="replay.gif",
        help="Where to save the produced GIF file.",
    )
    parser.add_argument(
        "--gif_duration",
        type=int,
        default=300,
        help="The duration of each gif image.",
    )
    flags = parser.parse_args()

    if flags.savedir == "args":
        flags.savedir = "{}_{}_{}.zip".format(
            time.strftime("%Y%m%d-%H%M%S"), flags.mode, flags.env
        )

    if flags.save_gif and "pixel_crop" not in flags.obs_keys:
        flags.obs_keys += ",pixel_crop"

    eval(**vars(flags))


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch.nn import functional as F


def compute_baseline_loss(advantages):
    return 0.5 * torch.sum(advantages ** 2)


def compute_entropy_loss(logits):
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    entropy_per_timestep = torch.sum(-policy * log_policy, dim=-1)
    return -torch.sum(entropy_per_timestep)


def compute_policy_gradient_loss(logits, actions, advantages):
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    policy_gradient_loss_per_timestep = cross_entropy * advantages.detach()
    return torch.sum(policy_gradient_loss_per_timestep)


def compute_forward_dynamics_loss(pred_next_emb, next_emb):
    forward_dynamics_loss = torch.norm(pred_next_emb - next_emb, dim=2, p=2)
    return torch.sum(torch.mean(forward_dynamics_loss, dim=1))


def compute_forward_binary_loss(pred_next_binary, next_binary):
    return F.binary_cross_entropy_with_logits(pred_next_binary, next_binary)


def compute_forward_class_loss(pred_next_glyphs, next_glyphs):
    next_glyphs = torch.flatten(next_glyphs, 0, 2).long()
    pred_next_glyphs = pred_next_glyphs.view(next_glyphs.size(0), -1)
    return F.cross_entropy(pred_next_glyphs, next_glyphs)


def compute_inverse_dynamics_loss(pred_actions, true_actions):
    inverse_dynamics_loss = F.cross_entropy(
        torch.flatten(pred_actions, 0, 1),
        torch.flatten(true_actions, 0, 1),
        reduction="none",
    )
    inverse_dynamics_loss = inverse_dynamics_loss.view_as(true_actions)
    return torch.sum(torch.mean(inverse_dynamics_loss, dim=1))



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation
from minihack.envs import register


class MiniHackHideAndSeekMapped(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        super().__init__(*args, des_file="hidenseek_mapped.des", **kwargs)


class MiniHackHideAndSeek(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        super().__init__(*args, des_file="hidenseek.des", **kwargs)


class MiniHackHideAndSeekLava(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        super().__init__(*args, des_file="hidenseek_lava.des", **kwargs)


class MiniHackHideAndSeekBig(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 400)
        super().__init__(*args, des_file="hidenseek_big.des", **kwargs)


register(
    id="MiniHack-HideNSeek-Mapped-v0",
    entry_point="minihack.envs.hidenseek:MiniHackHideAndSeekMapped",
)
register(
    id="MiniHack-HideNSeek-v0",
    entry_point="minihack.envs.hidenseek:MiniHackHideAndSeek",
)
register(
    id="MiniHack-HideNSeek-Lava-v0",
    entry_point="minihack.envs.hidenseek:MiniHackHideAndSeekLava",
)
register(
    id="MiniHack-HideNSeek-Big-v0",
    entry_point="minihack.envs.hidenseek:MiniHackHideAndSeekBig",
)



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import ast
import contextlib
import random
import os
import termios
import time
import timeit
import tty

import gym

import nle  # noqa: F401
import minihack  # noqa: F401
from nle import nethack
import tempfile
import shutil
import glob


_ACTIONS = tuple(
    [nethack.MiscAction.MORE]
    + list(nethack.CompassDirection)
    + list(nethack.CompassDirectionLonger)
)


@contextlib.contextmanager
def dummy_context():
    yield None


@contextlib.contextmanager
def no_echo():
    tt = termios.tcgetattr(0)
    try:
        tty.setraw(0)
        yield
    finally:
        termios.tcsetattr(0, termios.TCSAFLUSH, tt)


def get_action(env, action_mode, is_raw_env):
    if action_mode == "random":
        if not is_raw_env:
            action = env.action_space.sample()
        else:
            action = random.choice(_ACTIONS)
    elif action_mode == "human":
        while True:
            with no_echo():
                ch = ord(os.read(0, 1))
            if ch == nethack.C("c"):
                print("Received exit code {}. Aborting.".format(ch))
                return None
            try:
                if is_raw_env:
                    action = ch
                else:
                    action = env._actions.index(ch)
                break
            except ValueError:
                print(
                    (
                        "Selected action '%s' is not in action list. "
                        "Please try again."
                    )
                    % chr(ch)
                )
                continue
    return action


def play(
    env,
    mode,
    ngames,
    max_steps,
    seeds,
    savedir,
    no_render,
    render_mode,
    debug,
    save_gif,
    gif_path,
    gif_duration,
):
    env_name = env
    is_raw_env = env_name == "raw"

    if is_raw_env:
        if savedir is not None:
            os.makedirs(savedir, exist_ok=True)
            ttyrec = os.path.join(savedir, "nle.ttyrec.bz2")
        else:
            ttyrec = "/dev/null"
        env = nethack.Nethack(ttyrec=ttyrec)
    else:
        env_kwargs = dict(
            savedir=savedir,
            max_episode_steps=max_steps,
        )
        if save_gif:
            env_kwargs["observation_keys"] = ("pixel_crop",)
            try:
                import PIL.Image
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "To safe GIF files of trajectories, please install Pillow:"
                    " pip install Pillow"
                )

        env = gym.make(
            env_name,
            **env_kwargs,
        )
        if seeds is not None:
            env.seed(seeds)
        if not no_render:
            print("Available actions:", env._actions)

    obs = env.reset()

    steps = 0
    episodes = 0
    reward = 0.0
    action = None

    mean_sps = 0
    mean_reward = 0.0

    total_start_time = timeit.default_timer()
    start_time = total_start_time

    if save_gif:
        # Create a tmp directory for individual screenshots
        tmpdir = tempfile.mkdtemp()

    while True:
        if not no_render:
            if not is_raw_env:
                print("Previous reward:", reward)
                if action is not None:
                    print("Previous action: %s" % repr(env._actions[action]))
                env.render(render_mode)
            else:
                print("Previous action:", action)
                _, chars, _, _, blstats, message, *_ = obs
                msg = bytes(message)
                print(msg[: msg.index(b"\0")])
                for line in chars:
                    print(line.tobytes().decode("utf-8"))
                print(blstats)

        if save_gif:
            obs_image = PIL.Image.fromarray(obs["pixel_crop"])
            obs_image.save(os.path.join(tmpdir, f"e_{episodes}_s_{steps}.png"))

        action = get_action(env, mode, is_raw_env)
        if action is None:
            break

        if is_raw_env:
            obs, done = env.step(action)
        else:
            obs, reward, done, info = env.step(action)
        steps += 1

        if is_raw_env:
            done = done or steps >= max_steps  # NLE does this by default.
        else:
            mean_reward += (reward - mean_reward) / steps

        if not done:
            continue

        time_delta = timeit.default_timer() - start_time

        if not is_raw_env:
            print("Final reward:", reward)
            print("End status:", info["end_status"].name)
            print("Mean reward:", mean_reward)

        sps = steps / time_delta
        print("Episode: %i. Steps: %i. SPS: %f" % (episodes, steps, sps))

        episodes += 1
        mean_sps += (sps - mean_sps) / episodes

        start_time = timeit.default_timer()

        steps = 0
        mean_reward = 0.0

        if episodes == ngames:
            break
        env.reset()

    if save_gif:
        # Make the GIF and delete the temporary directory
        png_files = glob.glob(os.path.join(tmpdir, "e_*_s_*.png"))
        png_files.sort(key=os.path.getmtime)

        img, *imgs = [PIL.Image.open(f) for f in png_files]
        img.save(
            fp=gif_path,
            format="GIF",
            append_images=imgs,
            save_all=True,
            duration=gif_duration,
            loop=0,
        )
        shutil.rmtree(tmpdir)

        print("Saving replay GIF at {}".format(os.path.abspath(gif_path)))
    env.close()
    print(
        "Finished after %i episodes and %f seconds. Mean sps: %f"
        % (episodes, timeit.default_timer() - total_start_time, mean_sps)
    )


def main():
    parser = argparse.ArgumentParser(description="Play tool.")
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Enables debug mode, which will drop stack into "
        "an ipdb shell if an exception is raised.",
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=str,
        default="human",
        choices=["human", "random"],
        help="Control mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "-e",
        "--env",
        type=str,
        default="MiniHack-Room-Random-5x5-v0",
        help="Gym environment spec. Defaults to 'MiniHack-Room-Random-5x5-v0'.",
    )
    parser.add_argument(
        "-n",
        "--ngames",
        type=int,
        default=1,
        help="Number of games to be played before exiting. "
        "NetHack will auto-restart if > 1.",
    )
    parser.add_argument(
        "--max-steps",
        type=int,
        default=10000,
        help="Number of maximum steps per episode.",
    )
    parser.add_argument(
        "--seeds",
        default=None,
        help="Seeds to send to NetHack. Can be a dict or int. "
        "Defaults to None (no seeding).",
    )
    parser.add_argument(
        "--savedir",
        default="nle_data/play_data",
        help="Directory path where data will be saved. "
        "Defaults to 'nle_data/play_data'.",
    )
    parser.add_argument(
        "--no-render", action="store_true", help="Disables env.render()."
    )
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "full", "ansi"],
        help="Render mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "--save_gif",
        dest="save_gif",
        action="store_true",
        help="Saving a GIF replay of the evaluated episodes.",
    )
    parser.add_argument(
        "--no-save_gif",
        dest="save_gif",
        action="store_false",
        help="Do not save GIF.",
    )
    parser.set_defaults(save_gif=False)
    parser.add_argument(
        "--gif_path",
        type=str,
        default="replay.gif",
        help="Where to save the produced GIF file.",
    )
    parser.add_argument(
        "--gif_duration",
        type=int,
        default=300,
        help="The duration of each gif image.",
    )
    flags = parser.parse_args()

    if flags.debug:
        import ipdb

        cm = ipdb.launch_ipdb_on_exception
    else:
        cm = dummy_context

    with cm():
        if flags.seeds is not None:
            # to handle both int and dicts
            flags.seeds = ast.literal_eval(flags.seeds)

        if flags.savedir == "args":
            flags.savedir = "{}_{}_{}.zip".format(
                time.strftime("%Y%m%d-%H%M%S"), flags.mode, flags.env
            )

        play(**vars(flags))


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.

from minihack import MiniHack
from minihack.base import MH_DEFAULT_OBS_KEYS
from minihack.envs import register


class MiniHackSkill(MiniHack):
    """The base class for MiniHack Skill Acquisition tasks.

    Navigation tasks have the following characteristics:

    - The full action space is used.
    - Yes/No questions are enabled, but the menu-selection actions are disabled
      by default.
    - The character is set to a neutral human male caveman.
    - Maximum episode limit defaults to 250 (can be overriden via the
      `max_episode_steps` argument)
    - The default goal is to reach the stair down. This can be changed using
      a reward manager.
    - Auto-pick is disabled by default.
    - Inventory strings and corresponding letter are also included as part of
      the agent observations.
    """

    def __init__(
        self,
        *args,
        des_file,
        **kwargs,
    ):
        # Autopickup off by defautlt
        kwargs["autopickup"] = kwargs.pop("autopickup", False)
        # Allowing one-letter menu questions
        kwargs["allow_all_yn_questions"] = kwargs.pop(
            "allow_all_yn_questions", True
        )
        # Perform know steps
        kwargs["allow_all_modes"] = kwargs.pop("allow_all_modes", False)
        # Play with Caveman character by default
        kwargs["character"] = kwargs.pop("character", "cav-hum-new-mal")
        # Default episode limit
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 250)

        default_keys = MH_DEFAULT_OBS_KEYS + [
            "inv_strs",
            "inv_letters",
        ]
        kwargs["observation_keys"] = kwargs.pop(
            "observation_keys", default_keys
        )
        super().__init__(*args, des_file=des_file, **kwargs)


register(
    id="MiniHack-Skill-Custom-v0",
    entry_point="minihack.skills:MiniHackSkill",
)



# Copyright (c) Facebook, Inc. and its affiliates.
import gym
import minihack  # noqa: F401
import os
import numpy as np
import time
import argparse

try:
    from stable_baselines3.common.vec_env import DummyVecEnv, SubprocVecEnv
    from stable_baselines3.common.utils import set_random_seed
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use MiniGrid-based environments, please install"
        " stable_baselines3: pip install stable_baselines3"
    )


parser = argparse.ArgumentParser(description="Vectorized environment demo.")

parser.add_argument(
    "--env",
    type=str,
    default="MiniHack-MazeWalk-15x15-v0",
    help="MiniHack gym environment.",
)
parser.add_argument(
    "--num_env", type=int, default=4, help="Number of environments."
)
parser.add_argument(
    "--num_steps", type=int, default=10000, help="Number of environments."
)
parser.add_argument("--subproc", dest="subproc", action="store_true")
parser.add_argument("--no-subproc", dest="subproc", action="store_false")
parser.set_defaults(subproc=True)


class VecEnv_Wrapper:
    def __init__(self, env):
        self.env = env

    def step(self, action: int):
        os.chdir(self.env.env._vardir)
        return self.env.step(action)

    def reset(self):
        os.chdir(self.env.env._vardir)
        return self.env.reset()

    def close(self):
        os.chdir(self.env.env._vardir)
        self.env.close()

    def seed(self, core=None, disp=None, reseed=False):
        os.chdir(self.env.env._vardir)
        self.env.seed(core, disp, reseed)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(
                "attempted to get missing private attribute '{}'".format(name)
            )
        return getattr(self.env, name)


def make_env(env_id, subproc, rank, seed=0):
    """
    Utility function for multiprocessed env.

    :param env_id: (str) the environment ID
    :param num_env: (int) the number of environment you wish to have in subprocesses
    :param seed: (int) the inital seed for RNG
    :param rank: (int) index of the subprocess
    """

    def _init():
        env = gym.make(env_id)
        env.seed(seed + rank)
        if not subproc:
            env = VecEnv_Wrapper(env)
        return env

    set_random_seed(seed)
    return _init


def make_venv(args):
    if not args.subproc:
        # Performs actions sequentially
        venv = DummyVecEnv(
            [make_env(args.env, args.subproc, i) for i in range(args.num_env)]
        )
    else:
        # Performs actions in parallel processes
        venv = SubprocVecEnv(
            [make_env(args.env, args.subproc, i) for i in range(args.num_env)]
        )

    return venv


def main(args):
    env = make_venv(args)
    env.reset()

    start_time = time.time()
    for i in range(args.num_steps):
        env.step([np.random.randint(8)] * args.num_env)
        if (i - 1) % 200 == 0:
            env.reset()
    total_time_multi = time.time() - start_time

    print(
        "Took {:.2f}s with subproc={} on {} steps on {} envs - {:.2f} FPS".format(
            total_time_multi,
            args.subproc,
            args.num_steps,
            args.num_env,
            args.num_steps / total_time_multi,
        )
    )


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)



# Copyright (c) Facebook, Inc. and its affiliates.
from __future__ import annotations

import enum
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, List, Tuple

if TYPE_CHECKING:
    from minihack import MiniHack

from nle.nethack import Command, CompassDirection

Y_cmd = CompassDirection.NW


class EventType(enum.IntEnum):
    MESSAGE = 0
    LOC_ACTION = 1
    COORD = 2
    LOC = 3


COMESTIBLES = [
    "orange",
    "meatball",
    "meat ring",
    "meat stick",
    "kelp frond",
    "eucalyptus leaf",
    "clove of garlic",
    "sprig of wolfsbane",
    "carrot",
    "egg",
    "banana",
    "melon",
    "candy bar",
    "lump of royal jelly",
]


class Event(ABC):
    """An event which can occur in a MiniHack episode.

    This is the base class of all other events.
    """

    def __init__(
        self,
        reward: float,
        repeatable: bool,
        terminal_required: bool,
        terminal_sufficient: bool,
    ):
        """Initialise the Event.

        Args:
            reward (float):
                The reward for the event occuring
            repeatable (bool):
                Whether the event can occur repeated (i.e. if the reward can be
                collected repeatedly
            terminal_required (bool):
                Whether this event is required for the episode to terminate.
            terminal_sufficient (bool):
                Whether this event causes the episode to terminate on its own.
        """
        self.reward = reward
        self.repeatable = repeatable
        self.terminal_required = terminal_required
        self.terminal_sufficient = terminal_sufficient
        self.achieved = False

    @abstractmethod
    def check(self, env, previous_observation, action, observation) -> float:
        """Check whether the environment is in the state such that this event
        has occured.

        Args:
            env (MiniHack):
                The MiniHack environment in question.
            previous_observation (tuple):
                The previous state observation.
            action (int):
                The action taken.
            observation (tuple):
                The current observation.
        Returns:
            float: The reward.
        """
        pass

    def reset(self):
        """Reset the event, if there is any state necessary."""
        self.achieved = False

    def _set_achieved(self) -> float:
        if not self.repeatable:
            self.achieved = True
        return self.reward


def _standing_on_top(env, location):
    return not env.screen_contains(location)


class LocActionEvent(Event):
    """An event which checks whether an action is performed at a specified
    location.
    """

    def __init__(
        self,
        *args,
        loc: str,
        action: Command,
    ):
        """Initialise the Event.

        Args:
            loc (str):
                The name of the location to reach.
            action (int):
                The action to perform.
            reward (float):
                The reward for the event occuring
            repeatable (bool):
                Whether the event can occur repeated (i.e. if the reward can be
                collected repeatedly
            terminal_required (bool):
                Whether this event is required for the episode to terminate.
            terminal_sufficient (bool):
                Whether this event causes the episode to terminate on its own.
        """
        super().__init__(*args)
        self.loc = loc
        self.action = action
        self.status = False

    def check(self, env, previous_observation, action, observation) -> float:
        del previous_observation, observation
        if env._actions[action] == self.action and _standing_on_top(
            env, self.loc
        ):
            self.status = True
        elif env._actions[action] == Y_cmd and self.status:
            return self._set_achieved()
        else:
            self.status = False
        return 0

    def reset(self):
        super().reset()
        self.status = False


class LocEvent(Event):
    """An event which checks whether a specified location is reached."""

    def __init__(self, *args, loc: str):
        super().__init__(*args)
        """Initialise the Event.

        Args:
            loc (str):
                The name of the location to reach.
            reward (float):
                The reward for the event occuring
            repeatable (bool):
                Whether the event can occur repeated (i.e. if the reward can be
                collected repeatedly
            terminal_required (bool):
                Whether this event is required for the episode to terminate.
            terminal_sufficient (bool):
                Whether this event causes the episode to terminate on its own.
        """
        self.loc = loc

    def check(self, env, previous_observation, action, observation) -> float:
        del previous_observation, action, observation
        if _standing_on_top(env, self.loc):
            return self._set_achieved()
        return 0.0


class CoordEvent(Event):
    """An event which occurs when reaching certain coordinates."""

    def __init__(self, *args, coordinates: Tuple[int, int]):
        """Initialise the Event.

        Args:
            coordinates (tuple):
                The coordinates to reach for the event.
            reward (float):
                The reward for the event occuring
            repeatable (bool):
                Whether the event can occur repeated (i.e. if the reward can be
                collected repeatedly
            terminal_required (bool):
                Whether this event is required for the episode to terminate.
            terminal_sufficient (bool):
                Whether this event causes the episode to terminate on its own.
        """
        super().__init__(*args)
        self.coordinates = coordinates

    def check(self, env, previous_observation, action, observation) -> float:
        coordinates = tuple(observation[env._blstats_index][:2])
        if self.coordinates == coordinates:
            return self._set_achieved()
        return 0.0


class MessageEvent(Event):
    """An event which occurs when any of the `messages` appear."""

    def __init__(self, *args, messages: List[str]):
        """Initialise the Event.

        Args:
            messages (list):
                The messages to be seen to trigger the event.
            reward (float):
                The reward for the event occuring
            repeatable (bool):
                Whether the event can occur repeated (i.e. if the reward can be
                collected repeatedly
            terminal_required (bool):
                Whether this event is required for the episode to terminate.
            terminal_sufficient (bool):
                Whether this event causes the episode to terminate on its own.
        """
        super().__init__(*args)
        self.messages = messages

    def check(self, env, previous_observation, action, observation) -> float:
        del previous_observation, action
        curr_msg = (
            observation[env._original_observation_keys.index("message")]
            .tobytes()
            .decode("utf-8")
        )
        for msg in self.messages:
            if msg in curr_msg:
                return self._set_achieved()
        return 0.0


class AbstractRewardManager(ABC):
    """This is the abstract base class for the ``RewardManager`` that is used
    for defining custom reward functions.
    """

    def __init__(self):
        self.terminal_sufficient = None
        self.terminal_required = None

    @abstractmethod
    def collect_reward(self) -> float:
        """Return reward calculated and accumulated in check_episode_end_call,
        and then reset it.

        Returns:
            flaot: The reward.
        """
        raise NotImplementedError

    @abstractmethod
    def check_episode_end_call(
        self, env, previous_observation, action, observation
    ) -> bool:
        """Check if the task has ended, and accumulate any reward from the
        transition in ``self._reward``.

        Args:
            env (MiniHack):
                The MiniHack environment in question.
            previous_observation (tuple):
                The previous state observation.
            action (int):
                The action taken.
            observation (tuple):
                The current observation.
        Returns:
            bool: Boolean whether the episode has ended.

        """
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> None:
        """Reset all events, to be called when a new episode occurs."""
        raise NotImplementedError


class RewardManager(AbstractRewardManager):
    """This class is used for managing rewards, events and termination for
    MiniHack tasks.

    Some notes on the ordering or calls in the MiniHack/NetHack base class:

    - ``step(action)`` is called on the environment
    - Within ``step``, first a copy of the last observation is made, and then the
      underlying NetHack game is stepped
    - Then ``_is_episode_end(observation)`` is called to check whether this the
      episode has ended (and this is overridden if we've gone over our
      max_steps, or the underlying NetHack game says we're done (i.e. we died)
    - Then ``_reward_fn(last_observation, observation)`` is called to calculate
      the reward at this time-step
    - if ``end_status`` tells us the game is done, we quit the game
    - then ``step`` returns the observation, calculated reward, done, and some
    statistics.

    All this means that we need to check whether an observation is terminal in
    ``_is_episode_end`` before we're calculating the reward function.

    The call of ``_is_episode_end`` in ``MiniHack`` will call
    ``check_episode_end_call`` in this class, which checks for termination and
    accumulates any reward, which is returned and zeroed in ``collect_reward``.
    """

    def __init__(self):
        self.events: List[Event] = []
        self.custom_reward_functions: List[
            Callable[[MiniHack, Any, int, Any], float]
        ] = []
        self._reward = 0.0

        # Only used for GroupedRewardManager
        self.terminal_sufficient = None
        self.terminal_required = None

    def add_custom_reward_fn(
        self, reward_fn: Callable[[MiniHack, Any, int, Any], float]
    ) -> None:
        """Add a custom reward function which is called every after step to
        calculate reward.

        The function should be a callable which takes the environment, previous
        observation, action and current observation and returns a float reward.

        Args:
            reward_fn (Callable[[MiniHack, Any, int, Any], float]):
                A reward function which takes an environment, previous
                observation, action, next observation and returns a reward.

        """
        self.custom_reward_functions.append(reward_fn)

    def add_event(self, event: Event):
        """Add an event to be managed by the reward manager.

        Args:
            event (Event):
                The event to be added.
        """
        self.events.append(event)

    def _add_message_event(
        self, msgs, reward, repeatable, terminal_required, terminal_sufficient
    ):
        self.add_event(
            MessageEvent(
                reward,
                repeatable,
                terminal_required,
                terminal_sufficient,
                messages=msgs,
            )
        )

    def _add_loc_action_event(
        self,
        loc,
        action,
        reward,
        repeatable,
        terminal_required,
        terminal_sufficient,
    ):
        try:
            action = Command[action.upper()]
        except KeyError:
            raise KeyError(
                "Action {} is not in the action space.".format(action.upper())
            )

        self.add_event(
            LocActionEvent(
                reward,
                repeatable,
                terminal_required,
                terminal_sufficient,
                loc=loc.lower(),
                action=action,
            )
        )

    def add_eat_event(
        self,
        name: str,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add an event which is triggered when `name` is eaten.

        Args:
            name (str):
                The name of the object being eaten.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        msgs = [
            f"This {name} is delicious",
            "Blecch!  Rotten food!",
            "last bite of your meal",
        ]
        if name == "apple":
            msgs.append("Delicious!  Must be a Macintosh!")
            msgs.append("Core dumped.")
        if name == "pear":
            msgs.append("Core dumped.")

        self._add_message_event(
            msgs, reward, repeatable, terminal_required, terminal_sufficient
        )

    def add_wield_event(
        self,
        name: str,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered when a specific weapon is wielded.

        Args:
            name (str):
                The name of the weapon to be wielded.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        msgs = [
            f"{name} wields itself to your hand!",
            f"{name} (weapon in hand)",
        ]
        self._add_message_event(
            msgs, reward, repeatable, terminal_required, terminal_sufficient
        )

    def add_wear_event(
        self,
        name: str,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered when a specific armor is worn.

        Args:
            name (str):
                The name of the armor to be worn.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        msgs = [f"You are now wearing a {name}"]
        self._add_message_event(
            msgs, reward, repeatable, terminal_required, terminal_sufficient
        )

    def add_amulet_event(
        self,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered when an amulet is worn.

        Args:
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        self._add_message_event(
            ["amulet (being worn)."],
            reward,
            repeatable,
            terminal_required,
            terminal_sufficient,
        )

    def add_kill_event(
        self,
        name: str,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered when a specified monster is killed.

        Args:
            name (str):
                The name of the monster to be killed.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        self._add_message_event(
            [f"You kill the {name}"],
            reward,
            repeatable,
            terminal_required,
            terminal_sufficient,
        )

    def add_message_event(
        self,
        msgs: List[str],
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered when any of the given messages are seen.

        Args:
            msgs (List[str]):
                The name of the monster to be killed.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        self._add_message_event(
            msgs, reward, repeatable, terminal_required, terminal_sufficient
        )

    def add_positional_event(
        self,
        place_name: str,
        action_name: str,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered on taking a given action at a given place.

        Args:
            place_name (str):
                The name of the place to trigger the event.
            action_name (int):
                The name of the action to trigger the event.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        self._add_loc_action_event(
            place_name,
            action_name,
            reward,
            repeatable,
            terminal_required,
            terminal_sufficient,
        )

    def add_coordinate_event(
        self,
        coordinates: Tuple[int, int],
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered on when reaching the specified
        coordinates.

        Args:
            coordinates (Tuple[int, int]):
                The coordinates to be reached (tuple of ints).
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        self.add_event(
            CoordEvent(
                reward,
                repeatable,
                terminal_required,
                terminal_sufficient,
                coordinates=coordinates,
            )
        )

    def add_location_event(
        self,
        location: str,
        reward=1,
        repeatable=False,
        terminal_required=True,
        terminal_sufficient=False,
    ):
        """Add event which is triggered on reaching a specified location.

        Args:
            name (str):
                The name of the location to be reached.
            reward (float):
                The reward for this event. Defaults to 1.
            repeatable (bool):
                Whether this event can be triggered multiple times. Defaults to
                False.
            terminal_required (bool):
                Whether this event is required for termination. Defaults to
                True.
            terminal_sufficient (bool):
                Whether this event is sufficient for termination. Defaults to
                False.
        """
        self.add_event(
            LocEvent(
                reward,
                repeatable,
                terminal_required,
                terminal_sufficient,
                loc=location,
            )
        )

    def _set_achieved(self, event: Event) -> float:
        if not event.repeatable:
            event.achieved = True
        return event.reward

    def _standing_on_top(self, env, name):
        """Returns whether the agents is standing on top of the given object.
        The object name (e.g. altar, sink, fountain) must exist on the map.

        Args:
            env (MiniHack):
                The environment object.
            name (str):
                The name of the object.

        Returns:
            bool: True if the object name is not in the screen descriptions
            with agent info taking the space of the corresponding tile rather
            than the object).
        """
        return not env.screen_contains(name)

    def check_episode_end_call(
        self, env, previous_observation, action, observation
    ) -> bool:
        reward = 0.0
        for event in self.events:
            if event.achieved:
                continue
            reward += event.check(
                env, previous_observation, action, observation
            )

        for custom_reward_function in self.custom_reward_functions:
            reward += custom_reward_function(
                env, previous_observation, action, observation
            )
        self._reward += reward
        return self._check_complete()

    def _check_complete(self) -> bool:
        """Checks whether the episode is complete.

        Requires any event which is sufficient to be achieved, OR all required
        events to be achieved."""
        result = True
        for event in self.events:
            # This event is enough, we're done
            if event.achieved and event.terminal_sufficient:
                return True
            # We need this event and we haven't done it, we're not done
            if not event.achieved and event.terminal_required:
                result = False

        # We've achieved all terminal_required events, we're done
        return result

    def collect_reward(self) -> float:
        result = self._reward
        self._reward = 0.0
        return result

    def reset(self):
        self._reward = 0.0
        for event in self.events:
            event.reset()


class SequentialRewardManager(RewardManager):
    """A reward manager that ignores ``terminal_required`` and
    ``terminal_sufficient``, and just require every event is completed in the
    order it is added to the reward manager.
    """

    def __init__(self):
        self.current_event_idx = 0
        super().__init__()

    def check_episode_end_call(
        self, env, previous_observation, action, observation
    ):
        event = self.events[self.current_event_idx]
        reward = event.check(env, previous_observation, action, observation)
        if event.achieved:
            self.current_event_idx += 1
        self._reward += reward
        return self._check_complete()

    def _check_complete(self) -> bool:
        return self.current_event_idx == len(self.events)


class GroupedRewardManager(AbstractRewardManager):
    """Operates as a collection of reward managers.

    The rewards from each reward manager are summed, and termination can be
    specified by ``terminal_sufficient`` and ``terminal_required`` on each
    reward manager.

    Given this can be nested arbitrarily deeply (as each reward manager could
    itself be a GroupedRewardManager), this enables complex specification of
    groups of rewards.
    """

    def __init__(self):
        self.reward_managers: List[AbstractRewardManager] = []

    def check_episode_end_call(
        self, env, previous_observation, action, observation
    ) -> bool:
        for reward_manager in self.reward_managers:
            result = reward_manager.check_episode_end_call(
                env, previous_observation, action, observation
            )
            # This reward manager has completed and it's sufficient so we're
            # done
            if reward_manager.terminal_sufficient and result:
                return True
            # This reward manager is required and hasn't completed, so we're
            # not done
            if reward_manager.terminal_required and not result:
                return False

        # If we've got here we've completed all required reward managers, so
        # we're done
        return True

    def add_reward_manager(
        self,
        reward_manager: AbstractRewardManager,
        terminal_required: bool,
        terminal_sufficient: bool,
    ) -> None:
        """Add a new reward manager, with ``terminal_sufficient`` and
        ``terminal_required`` acting as for individual events.

        Args:
            reward_manager (RewardManager):
                The reward manager to be added.
            terminal_required (bool):
                Whether this reward manager terminating is required for the
                episode to terminate.
            terminal_sufficient:
                Whether this reward manager terminating is sufficient for the
                episode to terminate.
        """
        reward_manager.terminal_required = terminal_required
        reward_manager.terminal_sufficient = terminal_sufficient
        self.reward_managers.append(reward_manager)

    def collect_reward(self):
        reward = 0.0
        for reward_manager in self.reward_managers:
            reward += reward_manager.collect_reward()
        return reward

    def reset(self):
        self._reward = 0.0
        for reward_manager in self.reward_managers:
            reward_manager.reset()



# Copyright (c) Facebook, Inc. and its affiliates.
import gym
import numpy as np
import time
import minihack
import argparse
from minihack.agent.common.envs.wrapper import CachedEnvWrapper


def compare_speed(env, num_steps, queue_size):
    env_vanila = gym.make(env)
    test_speed(env_vanila, env, num_steps)
    env_vanila.close()

    env_queue = []
    for _ in range(queue_size):
        env_queue.append(gym.make(env))
    env_cached = CachedEnvWrapper(env_queue)
    test_speed(env_cached, env, num_steps)
    env_cached.close()


def test_speed(env, env_name, num_steps):
    """Tests the speed of an environment for num_steps steps."""
    start_time = time.time()
    env.reset()
    for _ in range(num_steps):
        _, _, done, _ = env.step(np.random.randint(8))
        if done:
            env.reset()
    total_time = time.time() - start_time

    print(
        "Took {:.4f}s to perform {} steps on {} envs - {:.2f} FPS".format(
            total_time, num_steps, env_name, num_steps / total_time
        )
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-e",
        "--env",
        type=str,
        default="MiniHack-MultiRoom-N2-Lava-v0",
        help="Gym environment spec. Defaults to 'MiniHack-MultiRoom-N2-Lava-v0'.",
    )
    parser.add_argument(
        "--num_steps",
        type=int,
        default=10000,
        help="Number of steps to run.",
    )
    parser.add_argument(
        "--queue_size",
        type=int,
        default=2,
        help="Number of environments in the queue.",
    )
    flags = parser.parse_args()

    compare_speed(**vars(flags))


if __name__ == "__main__":
    main()
